
#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python-Image-Bayerize - Converts an image to Bayerized form.
#   Copyright (C) 2014  Bill Smith
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from gimpfu import *
from array import array

def python_image_bayerize(img, layer, key):
    ''' This GIMP plugin constructs a Bayerized version of an image.

    Parameters:
    img    : The currently selected image.
    layer  : The layer of the currently selected image.
    key    : The colour filter array as string e.g. [BG][GR]
    '''
    precision = pdb.gimp_image_get_precision(img)
    bpp = layer.bpp
    if bpp > 8:
        gimp.message("Error - too many bytes per pixel")
        return
    width = layer.width
    height = layer.height
    progress = 0.0
    max_progress = float(width)
    gimp.progress_init("Bayerizing " + layer.name + "...")
    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)

    if key == 0:
        cfa = "BGGR"
    elif key == 1:
        cfa = "GBRG"
    elif key == 2:
        cfa = "GRBG"
    elif key == 3:
        cfa = "RGGB"

    home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
    home_pixels = array("B", home_rgn[0:width, 0:height])
    dest_rgn = layer.get_pixel_rgn(0, 0, width, height, True, True)
    dest_pixels = array("B", "\x00"*(width * height * bpp))
    pixel = array("B", "\x00"*bpp)
    newpix = array("B", "\x00"*bpp)

    for i in range(width):
        l = i-2*int(i/2)
        for j in range(height):
            n =  2*(j-2*int(j/2)) + l
            p = (i + width * j) * bpp
            pixel = home_pixels[p:p+bpp]
            if bpp <= 4:
                if cfa[n] == 'R':
                    newpix[0] = pixel[0]
                    newpix[1] = 0
                    newpix[2] = 0
                elif cfa[n] == 'G':
                    newpix[0] = 0
                    newpix[1] = pixel[1]
                    newpix[2] = 0
                else:
                    newpix[0] = 0
                    newpix[1] = 0
                    newpix[2] = pixel[2]
                if bpp > 3:
                    newpix[3] = pixel[3]
            elif bpp <= 8:
                if cfa[n] == 'R':
                    newpix[0] = pixel[0]
                    newpix[1] = pixel[1]
                    newpix[2] = 0
                    newpix[3] = 0
                    newpix[4] = 0
                    newpix[5] = 0
                elif cfa[n] == 'G':
                    newpix[0] = 0
                    newpix[1] = 0
                    newpix[2] = pixel[2]
                    newpix[3] = pixel[3]
                    newpix[4] = 0
                    newpix[5] = 0
                else:
                    newpix[0] = 0
                    newpix[1] = 0
                    newpix[2] = 0
                    newpix[3] = 0
                    newpix[4] = pixel[4]
                    newpix[5] = pixel[5]
                if bpp > 6:
                    newpix[6] = pixel[6]
                    newpix[7] = pixel[7]
            dest_pixels[p:p+bpp] = newpix
        progress += 1.0
        gimp.progress_update(progress / max_progress)

    dest_rgn[0:width, 0:height] = dest_pixels.tostring()
    layer.flush()
    layer.merge_shadow(True)
    layer.update(0, 0, width, height)
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()
    gimp.pdb.gimp_progress_end()

register(
    "python_image_bayerize",
    "Convert image to its Bayerized form",
    "Convert image to Bayerized form using (editable) colour filter array: [BG][GR]",
    "Bill Smith",
    "Bill Smith",
    "2014",
    "<Image>/Py-Astro/Colour Tools/_Bayerize image ...",
    "RGB*",
    [
        (PF_OPTION, "bayer_mask"  ,("Bayer mask: "), 0 ,["[BG][GR]", "[GB][RG]", "[GR][BG]", "[RG][GB]"]),
    ],
    [],
    python_image_bayerize,
)

main()
