#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python_Image_DeBayerize - de-Bayerizes a Bayerized image.
#   Copyright (C) 2014 Bill Smith
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from gimpfu import *
from array import array

def python_image_debayerize(img, layer, cfa):
    ''' This GIMP plugin de-Bayerizes a Bayerized image.

    Parameters:
    img    : The currently selected image.
    layer  : The layer of the currently selected image.
    cfa    : The colour filter array as string e.g. [BG][GR]
    '''

    precision = pdb.gimp_image_get_precision(img)
    bpp = layer.bpp
    if bpp > 8:
        gimp.message("Error - Too many bytes per pixel")
        return
    width = layer.width
    height = layer.height

    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)

    cfam=[0 for i in range(4)]
    if cfa == 0:
        cfam[0] = 0 # Blue
        cfam[1] = 1 # Green next to blue
        cfam[2] = 2 # Green next to red
        cfam[3] = 3 # Red
    elif cfa == 1:
        cfam[0] = 1
        cfam[1] = 0
        cfam[2] = 3
        cfam[3] = 2
    elif cfa == 2:
        cfam[0] = 2
        cfam[1] = 3
        cfam[2] = 0
        cfam[3] = 1
    elif cfa == 3:
        cfam[0] = 3
        cfam[1] = 2
        cfam[2] = 1
        cfam[3] = 0

    if not layer.is_rgb and not layer.is_grey:
        return

    home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
    home_pixels = array("B", home_rgn[0:width, 0:height])
    dest_rgn = layer.get_pixel_rgn(0, 0, width, height, True, True)
    dest_pixels = array("B", "\x00"*(width * height * bpp))

    progress = 0
    max_progress = float(height)
    gimp.progress_init("De-Bayerizing " + layer.name + "...")

    add = [0 for i in range(9)]
    pixout = array("B", "\x00"*bpp)
    pixel = [["\x00" for j in range(bpp)] for i in range(9)]

    for j in range(0, height):
        jm = j - 1
        jp = j + 1
        m = 2*(j-2*int(j/2))
        for i in range(0, width):
            im = i - 1
            ip = i + 1
            k = m+i-2*int(i/2)

            add[0] = pixAddr(im, jm, width, height, bpp)
            add[1] = pixAddr(i, jm, width, height, bpp)
            add[2] = pixAddr(ip, jm, width, height, bpp)
            add[3] = pixAddr(im, j, width, height, bpp)
            add[4] = pixAddr(i, j, width, height, bpp)
            add[5] = pixAddr(ip, j, width, height, bpp)
            add[6] = pixAddr(im, jp, width, height, bpp)
            add[7] = pixAddr(i, jp, width, height, bpp)
            add[8] = pixAddr(ip, jp, width, height, bpp)

            pixel[0] = home_pixels[add[0]:add[0]+bpp]
            pixel[1] = home_pixels[add[1]:add[1]+bpp]
            pixel[2] = home_pixels[add[2]:add[2]+bpp]
            pixel[3] = home_pixels[add[3]:add[3]+bpp]
            pixel[4] = home_pixels[add[4]:add[4]+bpp]
            pixel[5] = home_pixels[add[5]:add[5]+bpp]
            pixel[6] = home_pixels[add[6]:add[6]+bpp]
            pixel[7] = home_pixels[add[7]:add[7]+bpp]
            pixel[8] = home_pixels[add[8]:add[8]+bpp]

            if cfam[k] == 0:
                if bpp <= 4:
                    pixout[2] = pixel[4][2]
                    pixout[1] = ((pixel[1][1] + pixel[3][1] + pixel[5][1] + pixel[7][1]) >> 2)
                    pixout[0] = ((pixel[0][0] + pixel[2][0] + pixel[6][0] + pixel[8][0]) >> 2)
                elif bpp <= 8:
                    pixout[4] = pixel[4][4]
                    pixout[4] = pixel[4][5]
                    grn = ((pixel[1][2] + (pixel[1][3] << 8) + pixel[3][2] + (pixel[3][3] << 8) +
                    pixel[5][2] + (pixel[5][3] << 8) + pixel[7][2] + (pixel[7][3] << 8)) >> 2)
                    pixout[3] = (grn >> 8)
                    pixout[2] = grn - (pixout[3] << 8)
                    red = ((pixel[0][0] + (pixel[0][1] << 8) + pixel[2][0] + (pixel[2][1] << 8) +
                    pixel[6][0] + (pixel[6][1] << 8) + pixel[8][0] + (pixel[8][1] << 8)) >> 2)
                    pixout[1] =(red >> 8)
                    pixout[0] = red -  (pixout[1] << 8)
            elif cfam[k] == 1:
                if bpp <= 4:
                    pixout[1] = pixel[4][1]
                    pixout[2] = ((pixel[3][2] + pixel[5][2]) >> 1)
                    pixout[0] = ((pixel[1][0] + pixel[7][0]) >> 1)
                elif bpp <= 8:
                    pixout[2] = pixel[4][2]
                    pixout[3] = pixel[4][3]
                    blu = ((pixel[3][4] + (pixel[3][5] << 8) + pixel[5][4] + (pixel[5][5] << 8)) >> 1)
                    pixout[5] = (blu >> 8)
                    pixout[4] = blu - (pixout[5] << 8)
                    red = ((pixel[1][0] + (pixel[1][1] << 8) + pixel[7][0] + (pixel[7][1] << 8)) >> 1)
                    pixout[1] =(red >> 8)
                    pixout[0] = red -  (pixout[1] << 8)
            elif cfam[k] == 2:
                if bpp <= 4:
                    pixout[1] = pixel[4][1]
                    pixout[2] = ((pixel[1][2] + pixel[7][2]) >> 1)
                    pixout[0] = ((pixel[3][0] + pixel[5][0]) >> 1)
                elif bpp <= 8:
                    pixout[2] = pixel[4][2]
                    pixout[3] = pixel[4][3]
                    blu = ((pixel[1][4] + (pixel[1][5] << 8) + pixel[7][4] + (pixel[7][5] << 8)) >> 1)
                    pixout[5] = (blu >> 8)
                    pixout[4] = blu - (pixout[5] << 8)
                    red = ((pixel[3][0] + (pixel[3][1] << 8) + pixel[5][0] + (pixel[5][1] << 8)) >> 1)
                    pixout[1] =(red >> 8)
                    pixout[0] = red -  (pixout[1] << 8)
            elif cfam[k] == 3:
                if bpp <= 4:
                    pixout[0] = pixel[4][0]
                    pixout[1] = ((pixel[1][1] + pixel[3][1] + pixel[5][1] + pixel[7][1]) >> 2)
                    pixout[2] = ((pixel[0][2] + pixel[2][2] + pixel[6][2] + pixel[8][2]) >> 2)
                elif bpp <= 8:
                    pixout[0] = pixel[4][0]
                    pixout[1] = pixel[4][1]
                    grn = ((pixel[1][2] + (pixel[1][3] << 8) + pixel[3][2] + (pixel[3][3] << 8) +
                    pixel[5][2]+ (pixel[5][3] << 8) + pixel[7][2] + (pixel[7][3] << 8)) >> 2)
                    pixout[3] = (grn >> 8)
                    pixout[2] = grn - (pixout[3] << 8)
                    blu = ((pixel[0][4] + (pixel[0][5] << 8) + pixel[2][4] + (pixel[2][5] << 8) +
                    pixel[6][4] + (pixel[6][5] << 8) + pixel[8][4] + (pixel[8][5] << 8)) >> 2)
                    pixout[5] = (blu >> 8)
                    pixout[4] = blu - (pixout[5] << 8)
            if bpp == 4:
                pixout[3] = pixel[4][3]
            elif bpp == 8:
                    pixout[6] = pixel[4][6]
                    pixout[7] = pixel[4][7]
            dest_pixels[add[4]:add[4]+bpp] = pixout
        progress += 1.0
        gimp.progress_update(float(progress) / max_progress)

    dest_rgn[0:width, 0:height] = dest_pixels.tostring()
    layer.flush()
    layer.merge_shadow(True)
    layer.update(0, 0, width, height)
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()
    gimp.pdb.gimp_progress_end()

def pixAddr(x, y, width, height, bpp):
    if x < 0:
        x = -x
    elif x >= width:
        x = 2 * width -x - 2
    if y < 0:
        y = -y
    elif y >= height:
        y = 2 * height - y - 2
    return (x + width * y) * bpp

register(
    "python_image_debayerize",
    "De-Bayerize an image",
    "De-Bayerize an image using colour filter array of type: [BG][GR]",
    "Bill Smith",
    "Bill Smith",
    "2014",
    "<Image>/Py-Astro/Colour Tools/_De-Bayerize image ...",
    "RGB*",
    [
        (PF_OPTION, "bayer_mask"  ,("Bayer mask: "), 0 ,["[BG][GR]", "[GB][RG]", "[GR][BG]", "[RG][GB]"]),

    ],
    [],
    python_image_debayerize,
)

main()
