#!/usr/bin/env python
# Gimp-Python - allows the writing of Gimp plugins in Python.
#
# Python-Layer-Divide-Flat - divides a "flat" layer into all layers.
# Copyright (C) 2017  Bill Smith
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation and/or
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used
#    to endorse or promote products derived from this software without specific prior
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.

from gimpfu import *
from array import array

def python_layer_divide_flat(img, layer, flat):
    '''Divide a flat image into an image layer stack.

    Parameters:
    image     : Current image - home of the stacked images
    layer     : Layer of the current image (ignored).
    flat      : File name of flat image
    '''
    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)
    gimp.progress_init("Loading flat image file ...")
    try:
        if(flat.lower().endswith(('.jpeg'))):
            dimg = pdb.file_jpeg_load(flat, flat)
        elif(flat.lower().endswith(('.jpg'))):
            dimg = pdb.file_jpeg_load(flat, flat)
        elif(flat.lower().endswith(('.png'))):
            dimg = pdb.file_png_load(flat, flat)
        elif(flat.lower().endswith(('.tif'))):
            dimg = pdb.file_tiff_load(flat, flat)
        elif(flat.lower().endswith(('.tiff'))):
            dimg = pdb.file_tiff_load(flat, flat)
        elif(flat.lower().endswith(('.bmp'))):
            dimg = pdb.file_bmp_load(flat, flat)
        elif(flat.lower().endswith(('.gif'))):
            dimg = pdb.file_gif_load(flat, flat)
    except Exception as err:
        gimp.message("Error processing files: " + str(err))
    gimp.pdb.gimp_progress_end()

    if (dimg is None):
        gimp.message("The selected flat image file could not be opened.")
    else:
        flat_layer = dimg.layers[0]
        width = flat_layer.width
        height = flat_layer.height
        bpp = flat_layer.bpp
        if bpp != layer.bpp or width != layer.width or height != layer.height:
            gimp.message("Error - flat image and stack images not compatible")
        else:
            pixels=0.0
            kount=0.0
            percent=0.0
            median=0.0
            avg_red=0.0
            avg_grn=0.0
            avg_blu=0.0
            sig_red=0.0
            sig_grn=0.0
            sig_blu=0.0
            (avg_red,sig_red,median,pixels,kount,percent) = pdb.gimp_histogram(flat_layer,1,0,255)
            (avg_grn,sig_grn,median,pixels,kount,percent) = pdb.gimp_histogram(flat_layer,2,0,255)
            (avg_blu,sig_blu,median,pixels,kount,percent) = pdb.gimp_histogram(flat_layer,3,0,255)
            if bpp > 4:
                avg_red *= 65535.0
                avg_grn *= 65535.0
                avg_blu *= 65535.0
            flat_rgn = flat_layer.get_pixel_rgn(0, 0, width, height, False, False)
            flat_pixels = array("B", flat_rgn[0:width, 0:height])
            num = len(img.layers)
            last = width * height
            max_progress = float(last)
            for k in range(num):
                progress = 0.0
                layer = img.layers[num - k -1]
                gimp.progress_init("Dividing flat into " + layer.name + "...")
                home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
                home_pixels = array("B", home_rgn[0:width, 0:height])
                dest_rgn = layer.get_pixel_rgn(0, 0, width, height, True, True)
                dest_pixels = array("B", "\x00"*(width * height * bpp))
                i = 0
                n = 1
                for j in range(last):
                    dest_pixels[i] = zone8(home_pixels[i]*(avg_red/flat_pixels[i]))
                    dest_pixels[i+1] = zone8(home_pixels[i+1]*(avg_red/flat_pixels[i+1]))
                    dest_pixels[i+2] = zone8(home_pixels[i+2]*(avg_blu/flat_pixels[i+2]))
                    if bpp > 3:
                        dest_pixels[i+3] = home_pixels[i+3]
                    if (n == width):
                        n = 1
                        progress += float(width)
                        gimp.progress_update(progress / max_progress)
                    i += bpp
                    n += 1
                dest_rgn[0:width, 0:height] = dest_pixels.tostring()
                layer.flush()
                layer.merge_shadow(True)
                layer.update(0, 0, width, height)
                gimp.pdb.gimp_progress_end()
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()

def zone8(x):
#
# confine int(x) between 0 and 255
#
    if x < 0.0:
        p=0
    elif x > 255.0:
        p=255
    else:
        p=int(x)
    return p

def zone16(x):
#
# confine int(x) between 0 and 65535
#
    if x < 0.0:
        p=0
    elif x > 65535.0:
        p=65535
    else:
        p=int(x)
    return p

register(
    "python_layer_divide_flat",
    "Divide a flat image into a stack of images",
    "Divide a flat image into a stack of images in successive layers",
    "Bill Smith",
    "Bill Smith",
    "2017",
    "<Image>/Py-Astro/Layer Tools/_Divide layers by flat",
    "RGB*",
    [
        (PF_FILE, "file", "Select flat image file:", ""),
    ],
    [],
    python_layer_divide_flat,
)

main()
