
#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python-LRGB-Sharpen - Sharpens and colour enhances images.
#   Copyright (C) 2018  Bill Smith
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from gimpfu import *

def python_lrgb_sharpen(img, layer, desat, radius, strong, blur, bright):
    ''' This GIMP plugin sharpens an image using Okano-Dalby LRGB method.

    Parameters:
    img    : The currently selected image.
    layer  : The layer of the currently selected image.
    desat  : Desaturation option for luminance channel
    radius : Radius of unsharp mask gaussian blur
    strong : strength (opacity) of the unsharp mask
    blur   : Gaussian blurring of the colour layers
    bright : Brightness factor for colour layers
    '''

    precision = pdb.gimp_image_get_precision(img)
    bpp = layer.bpp
    if bpp> 8:
        gimp.message("Error - too many bytes per pixel")
        return
    key = [3, 2, 0, 1]
    gwdth = float(blur)
    desaturate = key[desat]
    value_mode = 14
    basec = 1.0 / bright
    divide_mode = 15
    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)

    layer1 = pdb.gimp_layer_copy(layer, 0)
    pdb.gimp_item_set_name(layer1, "Luminocity Mask")
    img.add_layer(layer1, -1)
    pdb.gimp_drawable_desaturate(layer1, desaturate)
    pdb.gimp_layer_set_mode(layer1, value_mode)
    pdb.plug_in_unsharp_mask(img, layer1, radius, strong, 255)
    if blur >= 0.1:
        pdb.plug_in_gauss_iir(img ,layer, blur, True, True)
    if bright >1.0:
        newlay = gimp.Layer(img, "Mask...", layer.width, layer.height, layer.type, 100.0 , divide_mode)
        gimp.set_background(basec, basec, basec)
        img.add_layer(newlay, -1)
        pdb.gimp_image_lower_item(img, newlay)
        pdb.gimp_drawable_fill(newlay, 1)
        pdb.gimp_image_merge_down(img, newlay, 0)
    pdb.gimp_image_merge_down(img, layer1, 0)

    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()

register(
    "python_lrgb_sharpen",
    "Sharpen image using LRGB method",
    "Sharpen image using LRGB method of Okano/Dalby",
    "Bill Smith",
    "Bill Smith",
    "2018",
    "<Image>/Py-Astro/Sharpen/_LRGB sharpen ...",
    "RGB*",
    [
        (PF_OPTION, "desat"  ,("Desaturation option: "), 0 ,[ "Luminance",
        "Average", "Lightness", "Luma"]),
        (PF_SLIDER, "radius", "Radius of unsharp mask", 2.0,(1.0, 10.0, 0.1)),
        (PF_SLIDER, "strong", "Strength of unsharp mask", 100.0,(50.0, 100.0, 1.0)),
        (PF_SLIDER, "blur"  , "Colour blur radius", 2.0,(0.0, 10.0, 0.1)),
        (PF_SLIDER, "bright", "Colour brightness factor", 1.0, (1.0, 2.0, 0.1)),
    ],
    [],
    python_lrgb_sharpen)

main()
