
#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python-Pick-Stars - Picks out the stars in an image.
#   Copyright (C) 2015  Bill Smith
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from gimpfu import *
from array import array

def python_pick_stars(img, layer, r_pick, g_pick, b_pick):
    ''' This GIMP plugin picks out stars in an image.

    Parameters:
    img    : The currently selected image.
    layer  : The layer of the currently selected image.
    r_pick : pick limit on colour red
    g_pick : pick limit on colour green
    b_pick : pick limit on colour blue
    '''
    precision = pdb.gimp_image_get_precision(img)
    bpp = layer.bpp
    if bpp <= 4:
        factor = 2.56
    elif bpp <= 8:
        factor = 655.36
    else:
        gimp.message("Error - too many bytes per pixel")
        return
    rmin=int(r_pick * factor)
    gmin=int(g_pick * factor)
    bmin=int(b_pick * factor)
    width = layer.width
    height = layer.height
    max_progress = float(height)
    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)
    pixel = array("B", "\x00"*bpp)
    newpix = array("B", "\x00"*bpp)
    progress = 0.0
    gimp.progress_init("Processing " + layer.name + "...")
    home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
    home_pixels = array("B", home_rgn[0:width, 0:height])
    dest_rgn = layer.get_pixel_rgn(0, 0, width, height, True, True)
    dest_pixels = array("B", "\x00"*(width * height * bpp))

    if bpp <= 4:
        for i in range(width):
            for j in range(height):
                p = (i + width * j) * bpp
                pixel = home_pixels[p:p+bpp]
                newpix[0] = pixel[0] if pixel[0] >= rmin else  0
                newpix[1] = pixel[1] if pixel[1] >= gmin else  0
                newpix[2] = pixel[2] if pixel[2] >= bmin else  0
                if bpp == 4:
                    newpix[3] = pixel[3]
                dest_pixels[p:p+bpp] = newpix
            progress += 1.0
            gimp.progress_update(progress / max_progress)
    elif bpp <= 8:
        for i in range(width):
            for j in range(height):
                p = (i + width * j) * bpp
                pixel = home_pixels[p:p+bpp]
                red = pixel[0] + (pixel[1] << 8)
                gre = pixel[2] + (pixel[3] << 8)
                blu = pixel[4] + (pixel[5] << 8)
                if red >= rmin:
                    newpix[0] = pixel[0]
                    newpix[1] = pixel[1]
                else:
                    newpix[0] = 0
                    newpix[1] = 0
                if gre >= gmin:
                    newpix[2] = pixel[2]
                    newpix[3] = pixel[3]
                else:
                    newpix[2] = 0
                    newpix[3] = 0
                if blu >= bmin:
                    newpix[4] = pixel[4]
                    newpix[5] = pixel[5]
                else:
                    newpix[4] = 0
                    newpix[5] = 0
                if bpp == 8:
                    newpix[6] = pixel[6]
                    newpix[7] = pixel[7]
                dest_pixels[p:p+bpp] = newpix
            progress += 1.0
            gimp.progress_update(progress / max_progress)
    dest_rgn[0:width, 0:height] = dest_pixels.tostring()
    layer.flush()
    layer.merge_shadow(True)
    layer.update(0, 0, width, height)
    gimp.pdb.gimp_progress_end()
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()

register(
    "python_pick_stars",
    "pick out the stars in an image",
    "pick out the stars above a max percentage brightness",
    "Bill Smith",
    "Bill Smith",
    "2014",
    "<Image>/Py-Astro/Colour Tools/_Pick out stars ...",
    "RGB*",
    [
        (PF_SLIDER, "r_pick", "Red limit", 90.0, (0.0, 100.0, 1.0)),
        (PF_SLIDER, "g_pick", "Green limit", 90.0, (0.0, 100.0, 1.0)),
        (PF_SLIDER, "b_pick", "Blue limit", 90.0, (0.0, 100.0, 1.0)),
    ],
    [],
    python_pick_stars,
)

main()
