
#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python-Resample-Image - Resamples image.
#   Copyright (C) 2015  Bill Smith
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from gimpfu import *
from array import array

def python_resample_image(img, layer, level):
    ''' This GIMP plugin resamples an image.

    Parameters:
    img    : The currently selected image.
    layer  : The layer of the currently selected image.
    level  : The resampling level.
    '''

    precision = pdb.gimp_image_get_precision(img)
    bpp = layer.bpp
    if bpp > 8:
        gimp.message("Error - too many bytes per pixel")
        return
    step = int(level)
    norm = float(step * step)
    width = layer.width
    height = layer.height
    width1 = step * (width / step)
    height1 = step * (height / step)
    progress = 0.0
    max_progress = float(width * height)
    gimp.progress_init("Resampling " + layer.name + "...")
    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)
    home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
    home_pixels = array("B", home_rgn[0:width, 0:height])
    dest_rgn = layer.get_pixel_rgn(0, 0, width, height, True, True)
    dest_pixels = array("B", "\x00"*(width * height * bpp))
    pixel = array("B", "\x00"*bpp)
    newpix = array("B", "\x00"*bpp)

    ii = 0
    for i in range(0, width1, step):
        jj = 0
        for j in range(0, height1, step):
            aaa = 0.0
            rrr = 0.5
            ggg = 0.5
            bbb = 0.5
            if bpp <= 4:
                for k in range(step):
                    for m in range(step):
                        p = (i + k + width * (j + m)) * bpp
                        pixel = home_pixels[p:p+bpp]
                        rrr += float(pixel[0])/norm
                        ggg += float(pixel[1])/norm
                        bbb += float(pixel[2])/norm
                        if bpp == 4:
                            aaa += float(pixel[3])/norm
                newpix[0] = min(int(rrr), 255)
                newpix[1] = min(int(ggg), 255)
                newpix[2] = min(int(bbb), 255)
                if bpp == 4:
                    newpix[3] = min(int(aaa), 255)
                q = (ii + width * jj) * bpp
                dest_pixels[q:q+bpp] = newpix
            elif bpp <=8:
                for k in range(step):
                    for m in range(step):
                        p = (i + k + width * (j + m)) * bpp
                        pixel = home_pixels[p:p+bpp]
                        rrr += float(pixel[0] + (pixel[1] << 8))/norm
                        ggg += float(pixel[2] + (pixel[3] << 8))/norm
                        bbb += float(pixel[4] + (pixel[5] << 8))/norm
                        if bpp == 8:
                            aaa += float(pixel[6] + (pixel[7] << 8))/norm
                rrr = min(rrr, 65535.0)
                ggg = min(ggg, 65535.0)
                bbb = min(bbb, 65535.0)
                nnn = int(rrr) >> 8
                newpix[0] = int(rrr) - (nnn << 8)
                newpix[1] = nnn
                nnn = int(ggg) >> 8
                newpix[2] = int(ggg) - (nnn << 8)
                newpix[3] = nnn
                nnn = int(bbb) >> 8
                newpix[4] = int(bbb) - (nnn << 8)
                newpix[5] = nnn
                if bpp == 8:
                    nnn = int(aaa) >> 8
                    newpix[6] = int(aaa) - (nnn << 8)
                    newpix[7] = nnn
                q = (ii + width * jj) * bpp
                dest_pixels[q:q+bpp] = newpix
            jj += 1
        ii += 1
        progress += float(width * step)
        gimp.progress_update(progress / max_progress)

    dest_rgn[0:width, 0:height] = dest_pixels.tostring()
    layer.flush()
    layer.merge_shadow(True)
    layer.update(0, 0, width, height)
    pdb.gimp_image_crop(img, width / step, height / step, 0, 0)
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()
    gimp.pdb.gimp_progress_end()

def min(a, b):
#
#   return the minimum of the numbers a and b
#
    if a <= b:
        n = a
    else:
        n = b
    return n

register(
    "python_resample_image",
    "Resample the image",
    "Resample the image over blocks of (level*level) pixels",
    "Bill Smith",
    "Bill Smith",
    "2015",
    "<Image>/Py-Astro/Misc Tools/_Resample image ...",
    "RGB*",
    [
        (PF_ADJUSTMENT, "level", "Resample level", 2, (2, 6, 1)),
    ],
    [],
    python_resample_image,
)

main()
