#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python_Round_Stars - make star images round!
#   Copyright (C) 2015 Bill Smith
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import math, colorsys
from gimpfu import *
from array import array

class star_chart:
#
#       star_chart class for finding stars in a photo
#       author - w.smith 2015
#
    def __init__(self, layer, minstar):
#
#   class for identifying stars in a photo
#   author = bill smith (2014)
#
        self.nstars = 0
        self.minstar = minstar
        self.layer = layer
        bpp = layer.bpp
        if bpp <= 4:
            star_factor = 2.56 * float(3 * minstar)
        elif bpp <= 8:
            star_factor = 655.36 * float(3 * minstar)
        maxpxl = 1000
        width = layer.width
        height = layer.height
        px = [0 for i in range(maxpxl)]
        py = [0 for i in range(maxpxl)]
        idp = [0 for i in range(maxpxl)]
        rgb = array("B", "\x00"*bpp)
        home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
        home_pixels = array("B", home_rgn[0:width, 0:height])
        progress = 0.0
        max_progress = float(width)
        gimp.progress_init("Finding star pixels...")

# find star pixels

        nnn = 0
        for i in range(width):
            for j in range(height):
                k = (i + width * j) * bpp
                rgb = home_pixels[k:k+bpp]
                if bpp <= 4:
                    starbright = float(rgb[0] + rgb[1] + rgb[2])
                elif bpp <= 8:
                    starbright = float((rgb[0] + (rgb[1] << 8)) + (rgb[2] + (rgb[3] << 8)) + (rgb[4] + (rgb[5] << 8)))
                if starbright >= star_factor:
                    px[nnn] = i
                    py[nnn] = j
                    idp[nnn] = nnn
                    nnn += 1
                    if nnn == maxpxl:
                        ee = [0 for n in range(100)]
                        px.extend(ee)
                        py.extend(ee)
                        idp.extend(ee)
                        maxpxl += 100
            progress += 1.0
            gimp.progress_update(progress / max_progress)
        gimp.pdb.gimp_progress_end()
        if nnn == 0:
            gimp.message("Warning - no stars found!")
            return

# group pixels into stars

        progress = 0
        max_progress = 0.5*float(nnn)*float(nnn-1)
        gimp.progress_init("Grouping pixels into stars...")
        self.nstars = nnn
        for i in range(nnn - 1):
            for j in range(i + 1, nnn):
                if abs(px[i] - px[j]) <= 1 and abs(py[i] - py[j]) <= 1:
                    if idp[j] == j:
                        idp[j] = idp[i]
                        self.nstars -= 1
                    elif idp[j] <> idp[i]:
                        p = min(idp[i], idp[j])
                        q = max(idp[i], idp[j])
                        for k in range(nnn):
                            if idp[k] == q:
                                idp[k] = p
                        self.nstars -= 1
                progress += 1.0
                if int(progress) % nnn == 0 :
                    gimp.progress_update(progress / max_progress)

        gimp.progress_update(1.0)
        gimp.pdb.gimp_progress_end()
        gimp.message("Number of stars found: "+str(self.nstars))

# sort pixels into star order

        gimp.progress_init("Sorting star pixels ...")
        lst = [i for i in range(nnn)]
        shell_sort(nnn, lst, idp)
        gimp.progress_update(1.0)
        gimp.pdb.gimp_progress_end()

# determine star properties: sizes and positions

        x1 = [0 for i in range(self.nstars)]
        x2 = [0 for i in range(self.nstars)]
        y1 = [0 for i in range(self.nstars)]
        y2 = [0 for i in range(self.nstars)]
        self.sa = [0 for i in range(self.nstars)]
        self.sx = [0 for i in range(self.nstars)]
        self.sy = [0 for i in range(self.nstars)]
        self.co = array("B","\x00"*bpp*self.nstars)

        progress = 1
        max_progress = float(nnn)
        gimp.progress_init("Calculating star properties...")

        sn = 0
        pop = 1
        me = idp[0]
        npx = lst[0]
        self.sx[sn] = px[npx]
        self.sy[sn] = py[npx]
        x1[sn] = px[npx]
        x2[sn] = px[npx]
        y1[sn] = py[npx]
        y2[sn] = py[npx]
        for i in range(1,nnn):
            npx = lst[i]
            if idp[i] == me:
                pop += 1
                self.sx[sn] += px[npx]
                self.sy[sn] += py[npx]
                x1[sn] = min(px[npx], x1[sn])
                x2[sn] = max(px[npx], x2[sn])
                y1[sn] = min(py[npx], y1[sn])
                y2[sn] = max(py[npx], y2[sn])
            else:
                self.sx[sn] /= pop
                self.sy[sn] /= pop
                wx = max(x2[sn]-self.sx[sn],self.sx[sn]-x1[sn])
                wy = max(y2[sn]-self.sy[sn],self.sy[sn]-y1[sn])
                self.sa[sn] = math.sqrt(wx*wx+wy*wy) + 1
                me = idp[i]
                sn += 1
                pop = 1
                self.sx[sn] = px[npx]
                self.sy[sn] = py[npx]
                x1[sn] = px[npx]
                x2[sn] = px[npx]
                y1[sn] = py[npx]
                y2[sn] = py[npx]
            progress += float(i)
            gimp.progress_update(progress / max_progress)

        self.sx[sn] /= pop
        self.sy[sn] /= pop
        wx = max(x2[sn]-self.sx[sn],self.sx[sn]-x1[sn])
        wy = max(y2[sn]-self.sy[sn],self.sy[sn]-y1[sn])
        self.sa[sn] = math.sqrt(wx*wx+wy*wy) + 1

        gimp.pdb.gimp_progress_end()

def shell_sort(n, lst, iii):
#
#       method to sort list of integers into ascending order
#       using the shell method
#       author - w.smith 2013
#
    m = n
    while m > 0:
        m = m / 2
        for j in range(n - m):
            i = j
            while i >= 0:
                l = i + m
                if iii[l] < iii[i]:
                    k = lst[i]
                    lst[i] = lst[l]
                    lst[l] = k
                    k = iii[i]
                    iii[i] = iii[l]
                    iii[l] = k
                    i -= m
                else:
                    i = -1

def python_round_stars(img, layer, arg0, arg1, arg2):
    ''' This GIMP plug-in makes star images round.

    Parameters:
    img       : The currently selected image.
    layer     : The layer of the currently selected image.
    minstar   : Minimum colour brightness of star objects
    scale     : Scale factor of star size

    '''
    minstar = int(arg0)
    scale = arg1
    npass = int(arg2)
    normal_mode = 0
    precision = pdb.gimp_image_get_precision(img)
    bpp = layer.bpp
    if bpp > 8:
        gimp.message("Error - too many bytes per pixel")
        return
    width = layer.width
    height = layer.height
    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)
    chart = star_chart(layer, minstar)
    home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
    home_pix = array("B", home_rgn[0:width, 0:height])

    for k in range(1, npass):
        progress = 0
        gimp.progress_init("Processing pass " + str(k) + " ...")
        coz = math.cos(2 * k * math.pi / npass)
        zin = math.sin(2 * k * math.pi / npass)
        opacity = 100.0 / float(k + 1)
        mask = gimp.Layer(img, "mask", width, height, layer.type, opacity, normal_mode)
        dest_rgn = mask.get_pixel_rgn(0, 0, width, height, True, True)
        dest_pix = array("B", home_rgn[0:width, 0:height])
        img.add_layer(mask, -1)

        for n in range(chart.nstars):
            m = int(chart.sa[n] * scale)
            for i in range(-m,m):
                for j in range(-m, m):
                    rr2 = float(i * i + j * j)
                    if rr2 <= float(m * m):
                        r = chart.sx[n] + i
                        s = chart.sy[n] + j
                        rs = (r + width * s) * bpp
                        u = chart.sx[n] + int(coz * i) - int(zin * j)
                        v = chart.sy[n] + int(zin * i) + int(coz * j)
                        if u >= 0 and u < width and v >= 0 and v < height:
                            uv = (u + width * v) * bpp
                            dest_pix[uv:uv+bpp] = home_pix[rs:rs+bpp]

            progress += 1
            gimp.progress_update(float(progress) / chart.nstars)
        dest_rgn[0:width, 0:height] = dest_pix.tostring()
        mask.flush()
        mask.merge_shadow(True)
        mask.update(0, 0, width, height)
        gimp.pdb.gimp_progress_end()

    img.flatten()
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()


def abs(x):
    if x < 0:
         x = -x
    return x

def min(x,y):
    if x < y:
        z = x
    else:
        z = y
    return z

def max(x,y):
    if x > y:
        z = x
    else:
        z = y
    return z

register(
    "python_round_stars",
    "Make star images round",
    "Make star images round",
    "Bill Smith",
    "Bill Smith",
    "2015",
    "<Image>/Py-Astro/Effects/_Round stars ...",
    "RGB, RGB*",
    [
        (PF_SLIDER, "minstar", "Star minimum", 85, (75, 100, 1)),
        (PF_SLIDER, "scale", "Star scale factor", 1.5, (1.0, 5.0, 0.1)),
        (PF_SLIDER, "npass", "Star roundness", 4, (4, 10, 1)),
    ],
    [],
    python_round_stars,
)

main()


