
#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python-Set-Dark-Sky - Set the image dark sky at chosen level.
#   Copyright (C) 2014  Bill Smith
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from gimpfu import *
from array import array
import random

def python_set_dark_sky(img, layer, olddark, newdark, noise, sigma):
    ''' This GIMP plugin sets the image dark sky.

    Parameters:
    img    : The currently selected image.
    layer  : The layer of the currently selected image (ignore).
    olddark: The current dark level
    newdark: The required sky dark point.
    noise  : Option to include random noise.
    sigma  : Root Mean Square of noise.
    '''
    precision = pdb.gimp_image_get_precision(img)
    bpp = layer.bpp
    if bpp <= 4:
        factor = 2.56
    elif bpp <= 8:
        factor = 655.36
    else:
        gimp.message("Error - too many bytes per pixel")
        return
    dark0=olddark * factor
    dark1=newdark * factor
    sigma=sigma * dark1 / 100.0
    width = layer.width
    height = layer.height
    bpp = layer.bpp
    max_progress = float(width)
    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)
    pixel = array("B", "\x00"*bpp)
    newpix = array("B", "\x00"*bpp)
    progress = 0.0
    gimp.progress_init("Processing " + layer.name + "...")
    home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
    home_pixels = array("B", home_rgn[0:width, 0:height])
    dest_rgn = layer.get_pixel_rgn(0, 0, width, height, True, True)
    dest_pixels = array("B", "\x00"*(width * height * bpp))
    for i in range(width):
        for j in range(height):
            p = (i + width * j) * bpp
            pixel = home_pixels[p:p+bpp]
            if noise:
                ru = max(random.gauss(dark1,sigma),0.0)
                gu = max(random.gauss(dark1,sigma),0.0)
                bu = max(random.gauss(dark1,sigma),0.0)
            else:
                ru = gu = bu = dark1
            if bpp <= 4:
                if pixel[0] <= int(dark0):
                    newpix[0] = int(ru)
                else:
                    newpix[0] = pixel[0]
                if pixel[1] <= int(dark0):
                    newpix[1] = int(gu)
                else:
                    newpix[1] = pixel[1]
                if pixel[2] <= int(dark0):
                    newpix[2] = int(bu)
                else:
                    newpix[2] = pixel[2]
                if bpp == 4:
                    newpix[3] = pixel[3]
            elif bpp <= 8:
                if (pixel[0] + (pixel[1] << 8)) <= int(dark0):
                    rr = int(ru)
                    newpix[1] = (rr >> 8)
                    newpix[0] = rr - (newpix[1] << 8)
                else:
                    newpix[0] = pixel[0]
                    newpix[1] = pixel[1]
                if (pixel[2] + (pixel[3] << 8)) <= int(dark0):
                    gg = int(gu)
                    newpix[3] = (gg >> 8)
                    newpix[2] = gg - (newpix[3] << 8)
                else:
                    newpix[2] = pixel[2]
                    newpix[3] = pixel[3]
                if (pixel[4] + (pixel[5] << 8)) <= int(dark0):
                    bb = int(bu)
                    newpix[5] = (bb >> 8)
                    newpix[4] = bb - (newpix[5] << 8)
                else:
                    newpix[4] = pixel[4]
                    newpix[5] = pixel[5]
                if bpp == 8:
                    newpix[6] = pixel[6]
                    newpix[7] = pixel[7]
            dest_pixels[p:p+bpp] = newpix
        progress += 1.0
        gimp.progress_update(progress / max_progress)
    dest_rgn[0:width, 0:height] = dest_pixels.tostring()
    layer.flush()
    layer.merge_shadow(True)
    layer.update(0, 0, width, height)
    gimp.pdb.gimp_progress_end()
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()

def max(x, y):
#
#   return maximum of numbers x and y
#
    if x >= y:
        p = x
    else:
        p = y
    return p

register(
    "python_set_dark_sky",
    "Set the dark sky",
    "Set the image dark sky with optional random noise",
    "Bill Smith",
    "Bill Smith",
    "2014",
    "<Image>/Py-Astro/Effects/_Set image dark sky ...",
    "RGB*",
    [
        (PF_SLIDER, "olddark", "Current sky dark level (%)", 25.0, (0.0, 100.0, 1.0)),
        (PF_SLIDER, "newdark", "Set new sky dark level (%)", 15.0, (0.0, 100.0, 1.0)),
        (PF_TOGGLE, "noise", "Add random noise", TRUE),
        (PF_SLIDER, "sigma", "RMS noise level", 10.0, (0.0, 25.0, 1.0)),
    ],
    [],
    python_set_dark_sky,
)

main()
