
#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python-Neutral-Dark-Sky - Set the image sky colour to neutral.
#   Copyright (C) 2015  Bill Smith
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from gimpfu import *
from array import array
import random

def python_set_neutral_sky(img, layer):
    ''' This GIMP plugin sets the image sky to a neutral colour.

    Parameters:
    img    : The currently selected image.
    layer  : The layer of the currently selected image.
    '''
    precision = pdb.gimp_image_get_precision(img)
    bpp = layer.bpp
    if bpp > 8:
        gimp.message("Error - too many bytes per pixel")
        return
    num=0.0
    pixels=0.0
    kount=0.0
    percent=0.0
    median=0.0
    avg_red=0.0
    avg_grn=0.0
    avg_blu=0.0
    flc_red=0.0
    flc_grn=0.0
    flc_blu=0.0
    (avg_red,flc_red,median,pixels,kount,percent) = pdb.gimp_histogram(layer,1,0,255)
    (avg_grn,flc_grn,median,pixels,kount,percent) = pdb.gimp_histogram(layer,2,0,255)
    (avg_blu,flc_blu,median,pixels,kount,percent) = pdb.gimp_histogram(layer,3,0,255)
    if bpp > 4:
        avg_red *= 65535.0
        avg_grn *= 65535.0
        avg_blu *= 65535.0
        flc_red *= 65535.0
        flc_grn *= 65535.0
        flc_blu *= 65535.0
    if flc_red < 1.0 or flc_grn < 1.0 or flc_blu < 1.0:
        gimp.message("Error - Negligible colour variance detected")
        return
    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)
    progress=0.0
    width = layer.width
    height = layer.height
    max_progress = float(width)
    ffr=(flc_grn/flc_red)
    ffb=(flc_grn/flc_blu)
    ccr=ffr*avg_red-avg_grn
    ccb=ffb*avg_blu-avg_grn
    gimp.progress_init("Correcting sky colour ...")
    pixel = array("B", "\x00"*bpp)
    newpix = array("B", "\x00"*bpp)
    home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
    home_pixels = array("B", home_rgn[0:width, 0:height])
    dest_rgn = layer.get_pixel_rgn(0, 0, width, height, True, True)
    dest_pixels = array("B", "\x00"*(width * height * bpp))
    for i in range(width):
        for j in range(height):
            p = (i + width * j) * bpp
            pixel = home_pixels[p:p+bpp]
            if bpp <= 4:
                newpix[0] = zone8(ffr*pixel[0]-ccr)
                newpix[1] = pixel[1]
                newpix[2] = zone8(ffb*pixel[2]-ccb)
                if bpp > 3:
                    newpix[3]=pixel[3]
            elif bpp <= 8:
                rrr = zone16(ffr*(pixel[0]+(pixel[1] << 8))-ccr)
                newpix[1] = (rrr >> 8)
                newpix[0] = rrr - (newpix[1] << 8)
                ggg = pixel[2]+(pixel[3] << 8)
                newpix[3] = (ggg >> 8)
                newpix[2] = ggg - (newpix[3] << 8)
                bbb = zone16(ffb*(pixel[4]+(pixel[5] << 8))-ccb)
                newpix[5] = (bbb >> 8)
                newpix[4] = bbb - (newpix[5] << 8)
                if bpp > 6:
                    newpix[6]=pixel[6]
                    newpix[7]=pixel[7]
            dest_pixels[p:p+bpp] = newpix
        progress += 1.0
        gimp.progress_update(progress / max_progress)
    dest_rgn[0:width, 0:height] = dest_pixels.tostring()
    layer.flush()
    layer.merge_shadow(True)
    layer.update(0, 0, width, height)
    gimp.pdb.gimp_progress_end()
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()

def zone8(x):
#
# confine int(x) between 0 and 255
#
    if x < 0.0:
        p=0
    elif x > 255.0:
        p=255
    else:
        p=int(x)
    return p

def zone16(x):
#
# confine int(x) between 0 and 65535
#
    if x < 0.0:
        p=0
    elif x > 65535.0:
        p=65535
    else:
        p=int(x)
    return p

register(
    "python_set_neutral_sky",
    "Set the sky colour to neutral",
    "Set the image sky colour to neutral",
    "Bill Smith",
    "Bill Smith",
    "2015",
    "<Image>/Py-Astro/Colour Tools/_Neutralise sky colour",
    "RGB*",
    [
    ],
    [],
    python_set_neutral_sky,
)

main()
