
#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python-Enhance-Using SVD layers - Boosts the detail in star images.
#   Copyright (C) 2016  Bill Smith
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from gimpfu import *
from array import array

def python_smart_sharpen(img, layer, edgrad, gaurad, shprad, amount, thresh):
    ''' This GIMP plugin sharpens an image using smart sharpening.

    Parameters:
    img     : The currently selected image.
    layer   : The layer of the currently selected image.
    '''

    # set up

    bpp = layer.bpp
    if bpp> 8:
        gimp.message("Error - too many bytes per pixel")
        return
    showme = False
    width = layer.width
    height = layer.height
    title = pdb.gimp_item_get_name(layer)
    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)

    # construct the edges from a duplicate image

    dup = pdb.gimp_image_duplicate(img)
    duplay = pdb.gimp_image_get_active_layer(dup)
    pdb.plug_in_edge(dup, duplay, edgrad, 1, 0)
    pdb.gimp_image_convert_grayscale(dup)
    pdb.gimp_drawable_levels(duplay, 0, 0.2, 0.8, True, 1.0,  0.0, 1.0, True)
    pdb.plug_in_gauss_iir(dup, duplay, gaurad, True, True)
    pdb.gimp_drawable_levels(duplay, 0, 0.0, 0.9, True, 1.0,  0.0, 1.0, True)
    if showme:
        pdb.gimp_item_set_name(duplay, "Sharpen Mask")
        disp1 = gimp.Display(dup)

    # perform LAB colour decomposition

    TUP = pdb.plug_in_decompose(img, layer, "LAB", 1)
    lab = TUP[0]
    if showme:
        disp2 = gimp.Display(lab)

    # isolate LAB layers and blur colour layers

    lll = pdb.gimp_image_get_layer_by_name(lab,'L')
    aaa = pdb.gimp_image_get_layer_by_name(lab,'A')
    pdb.plug_in_gauss_iir(lab, aaa, gaurad, True, True)
    bbb = pdb.gimp_image_get_layer_by_name(lab,'B')
    pdb.plug_in_gauss_iir(lab, bbb, gaurad, True, True)
    pdb.gimp_image_set_active_layer(lab, lll)

    # add the edges image as a channel selection

    mask = pdb.gimp_channel_new_from_component(lab, GRAY_CHANNEL, "Sharpen Mask")
    dummy = pdb.gimp_image_get_active_channel(lab)
    pdb.gimp_image_insert_channel(lab, mask, dummy, -1)
    pdb.gimp_edit_copy(duplay)
    floater = pdb.gimp_edit_paste(mask, True)
    pdb.gimp_floating_sel_anchor(floater)
    pdb.gimp_item_set_visible(mask, True)
    pdb.gimp_image_set_active_layer(lab, lll)

    # sharpen the luminosity layer

    pdb.plug_in_unsharp_mask(lab, lll, shprad, amount, thresh)
    #pdb.plug_in_sharpen(lab, lll, 50.0)

    # compose the final colour image

    null = gimp.Layer(lab, "dummy", width, height,
                          GRAY_IMAGE, 100, NORMAL_MODE)
    new = pdb.plug_in_drawable_compose(lab, lll, aaa, bbb, null, "LAB")

    if showme:
        disp3 = gimp.Display(new)

    # add new image as layer on top of original image (for comparison)

    newlay = pdb.gimp_image_get_active_layer(new)
    pdb.gimp_edit_copy(newlay)
    floater = pdb.gimp_edit_paste(layer, True)
    pdb.gimp_floating_sel_to_layer(floater)
    top = pdb.gimp_image_get_active_layer(img)
    pdb.gimp_item_set_name(top, "SHP_" + title)
    pdb.gimp_image_merge_down(img, top, 0)

    # clean up

    if showme:
        pdb.gimp_display_delete(disp1)
        pdb.gimp_display_delete(disp2)
        pdb.gimp_display_delete(disp3)
    else:
        pdb.gimp_image_delete(lab)
        pdb.gimp_image_delete(dup)
        pdb.gimp_image_delete(new)

    # wind up

    #gimp.message("Procedure ended correctly")
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()


register(
    "python_smart_sharpen",
    "Smart sharpen",
    "Smart sharpen using edge detection",
    "Bill Smith 2018",
    "Bill Smith 2018",
    "2018",
    "<Image>/Py-Astro/Sharpen/_Smart sharpen ...",
    "RGB*",
    [
        (PF_SLIDER, "edgrad", "Edge width", 3.0, (1.0, 12.0, 0.1)),
        (PF_SLIDER, "gaurad", "Gaussian blur radius", 3.0, (1.0, 10.0, 0.1)),
        (PF_SLIDER, "shprad", "Sharpen radius", 2.0, (1.0, 10.0, 0.1)),
        (PF_SLIDER, "amount", "Sharpen amount", 0.5, (0.1, 1.0, 0.1)),
        (PF_SLIDER, "thresh", "Sharpen threshold", 20, (0, 255, 1)),
    ],
    [],
    python_smart_sharpen)

main()
