
#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python-Split-Channels - Splits the image into colour channels.
#   Copyright (C) 2015  Bill Smith
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from gimpfu import *
from array import array

def python_split_channels(img, layer):
    ''' This GIMP plugin splits an image into its colour channels as layers.

    Parameters:
    img    : The currently selected image.
    layer  : The layer of the currently selected image.
    '''

    precision = pdb.gimp_image_get_precision(img)
    bpp = layer.bpp
    if bpp > 8:
        gimp.message("Error - too many bytes per pixel")
        return
    width = layer.width
    height = layer.height
    max_progress = float(width)
    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)
    pixel = array("B", "\x00"*bpp)
    newpix_r = array("B", "\x00"*bpp)
    newpix_g = array("B", "\x00"*bpp)
    newpix_b = array("B", "\x00"*bpp)
    progress = 0.0
    gimp.progress_init("Splitting channels of " + layer.name + "...")
    red_layer = gimp.Layer(img, "red channel", width, height, layer.type, layer.opacity, layer.mode)
    grn_layer = gimp.Layer(img, "green channel", width, height, layer.type, layer.opacity, layer.mode)
    blu_layer = gimp.Layer(img, "blue channel", width, height, layer.type, layer.opacity, layer.mode)
    home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
    home_pixels = array("B", home_rgn[0:width, 0:height])
    dest_rgn_r = red_layer.get_pixel_rgn(0, 0, width, height, True, True)
    dest_pixels_r = array("B", "\x00"*(width * height * bpp))
    dest_rgn_g = grn_layer.get_pixel_rgn(0, 0, width, height, True, True)
    dest_pixels_g = array("B", "\x00"*(width * height * bpp))
    dest_rgn_b = blu_layer.get_pixel_rgn(0, 0, width, height, True, True)
    dest_pixels_b = array("B", "\x00"*(width * height * bpp))
    if bpp <= 4:
        for i in range(width):
            for j in range(height):
                p = (i + width * j) * bpp
                pixel = home_pixels[p:p+bpp]
                newpix_r[0] = pixel[0]
                newpix_g[1] = pixel[1]
                newpix_b[2] = pixel[2]
                if bpp == 4:
                    newpix_r[3] = pixel[3]
                    newpix_g[3] = pixel[3]
                    newpix_b[3] = pixel[3]
                dest_pixels_r[p:p+bpp] = newpix_r
                dest_pixels_g[p:p+bpp] = newpix_g
                dest_pixels_b[p:p+bpp] = newpix_b
            progress += 1.0
            gimp.progress_update(progress / max_progress)
    elif bpp <= 8:
        for i in range(width):
            for j in range(height):
                p = (i + width * j) * bpp
                pixel = home_pixels[p:p+bpp]
                newpix_r[0] = pixel[0]
                newpix_r[1] = pixel[1]
                newpix_g[2] = pixel[2]
                newpix_g[3] = pixel[3]
                newpix_b[4] = pixel[4]
                newpix_b[5] = pixel[5]
                if bpp == 8:
                    newpix_r[6] = pixel[6]
                    newpix_r[7] = pixel[7]
                    newpix_g[6] = pixel[6]
                    newpix_g[7] = pixel[7]
                    newpix_b[6] = pixel[6]
                    newpix_b[7] = pixel[7]
                dest_pixels_r[p:p+bpp] = newpix_r
                dest_pixels_g[p:p+bpp] = newpix_g
                dest_pixels_b[p:p+bpp] = newpix_b
            progress += 1.0
            gimp.progress_update(progress / max_progress)
    dest_rgn_r[0:width, 0:height] = dest_pixels_r.tostring()
    dest_rgn_g[0:width, 0:height] = dest_pixels_g.tostring()
    dest_rgn_b[0:width, 0:height] = dest_pixels_b.tostring()
    img.add_layer(blu_layer, -1)
    img.add_layer(grn_layer, -1)
    img.add_layer(red_layer, -1)
    img.remove_layer(layer)
    red_layer.flush()
    red_layer.merge_shadow(True)
    red_layer.update(0, 0, width, height)
    grn_layer.flush()
    grn_layer.merge_shadow(True)
    grn_layer.update(0, 0, width, height)
    blu_layer.flush()
    blu_layer.merge_shadow(True)
    blu_layer.update(0, 0, width, height)
    gimp.pdb.gimp_progress_end()
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()

register(
    "python_split_channels",
    "Split image into colour channels",
    "Split image into colour channels",
    "Bill Smith",
    "Bill Smith",
    "2015",
    "<Image>/Py-Astro/Colour Tools/_Split colour channels",
    "RGB*",
    [],
    [],
    python_split_channels
)

main()
