#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python_Stack_Image_Layers - aligns and stacks #images
#   Copyright (C) 2016 Bill Smith
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import math
from gimpfu import *
from array import array

class calibrate_image:
#
#       Calibrate an image layer
#       author - w.smith 2014
#
    def __init__(self, layer, minbright):
#
#   class for calibrating a photo
#   author = bill smith (2016)
#
        self.layer = layer
        bpp = layer.bpp
        if bpp > 8:
            gimp.message("Error - too many bytes per pixel")
            return
        width = layer.width
        height = layer.height
        self.sx = [0 for i in range(width)]
        self.sy = [0 for i in range(height)]
        rwidth = 1.0 / width
        rheight = 1.0 /height
        rgb = array("B", "\x00"*bpp)
        home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
        home_pixels = array("B", home_rgn[0:width, 0:height])
        progress = 0.0
        max_progress = float(width)
        gimp.progress_init("Analysing image " + layer.name + "...")

# construct data arrays

        for i in range(width):
            for j in range(height):
                k = (i + width * j) * bpp
                rgb = home_pixels[k:k+bpp]
                if bpp <= 4:
                    bright = float(rgb[0] + rgb[1] + rgb[2]) / 768.0
                elif bpp <= 8:
                    bright = float((rgb[0] + (rgb[1] << 8)) + (rgb[2] + (rgb[3] << 8)) + (rgb[4] + (rgb[5] << 8))) / 196608.0
                if bright > minbright:
                    self.sx[i] += (bright * rheight)
                    self.sy[j] += (bright * rwidth)
            progress += 1.0
            gimp.progress_update(progress / max_progress)
        gimp.pdb.gimp_progress_end()

def python_stack_image_layers(img, layer, arg0, arg1, mode, pack, auto):
    ''' This GIMP plug-in aligns and stacks star images.

    Parameters:
    img        : The currently selected image.
    ref_layer  : The selected (refererence) layer of the current image.
    min_bright : Minimum selected pixel brightness.
    search_box : Width of search box for fitting (in pixels).
    mode       : Blending mode for layers.
    pack       : Flatten final image option.
    auto       : Autocrop finalimage option.

    '''
    key = [0, 7, 4]
    if not pack:
        auto = False
    min_bright = int(arg0) / 100.0
    half_box = (int(arg1) / 2)
    search_box = 2 * (int(arg1) /2) +1
    ref_layer = layer
    width = ref_layer.width
    height = ref_layer.height
    num = len(img.layers)
    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)
    pdb.gimp_item_set_visible(ref_layer,True)
    maxprog = float(search_box)
    ref_calib = calibrate_image(ref_layer, min_bright)
    xpos = ypos = xneg = yneg = 0

    for k in range(num):
        mov_layer = img.layers[num - k - 1]
        chkvis = pdb.gimp_item_get_visible(img.layers[num - k - 1])
        if(chkvis == True and mov_layer != ref_layer):
            progress = 0
            mov_calib = calibrate_image(mov_layer, min_bright)
            gimp.progress_init("Aligning image " + mov_layer.name + " ...")
            xdis = 0
            ydis = 0
            overmax_x = 0.0
            overmax_y = 0.0
            for n in range(-half_box,half_box+1):
                overlap_x = 0.0
                overlap_y = 0.0
                progress += 1.0
                gimp.progress_update(progress / maxprog)
                if n < 0:
                    for i in range(width-half_box):
                        overlap_x += ref_calib.sx[i] * mov_calib.sx[i-n]
                    for i in range(height-half_box):
                        overlap_y += ref_calib.sy[i] * mov_calib.sy[i-n]
                else:
                    for i in range(width-half_box):
                        overlap_x += ref_calib.sx[i+n] * mov_calib.sx[i]
                    for i in range(height-half_box):
                        overlap_y += ref_calib.sy[i+n] * mov_calib.sy[i]
                if overlap_x > overmax_x:
                    xdis = n
                    overmax_x = overlap_x
                if overlap_y > overmax_y:
                    ydis = n
                    overmax_y = overlap_y
            if xdis <> 0 or ydis <> 0:
                pdb.gimp_layer_translate (mov_layer, xdis, ydis)
                if auto:
                    xpos = max(xpos, xdis)
                    ypos = max(ypos, ydis)
                    xneg = min(xneg, xdis)
                    yneg = min(yneg, ydis)
            gimp.pdb.gimp_progress_end()
            gimp.message("Displacement of "+mov_layer.name+":   "+str(xdis)+"   "+str(ydis))

    if pack:
        n = 0
        for i in range(num):
            if (pdb.gimp_item_get_visible(img.layers[i]) == True):
                n += 1
        k = 0
        for i in range(num):
            if (pdb.gimp_item_get_visible(img.layers[i]) == True):
                if (key[mode] == 0):
                    blend = int(1000 / (n - k)) / 10.0
                else :
                    blend = 100
                pdb.gimp_layer_set_opacity(img.layers[i], blend)
                pdb.gimp_layer_set_mode(img.layers[i], key[mode])
                k += 1
        gimp.progress_init("Flattening image ...")
        img.flatten()
        gimp.pdb.gimp_progress_end()
        if auto:
            new_width = width - xpos + xneg
            new_height = height - ypos + yneg
            pdb.gimp_image_crop(img, new_width, new_height, xpos, ypos)
        pdb.gimp_item_set_name(img.layers[0],"Stacked image")
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()

register(
    "python_stack_image_layers",
    "Align and stack image layers",
    "Align and stack image layers - shifting only (no rotation)",
    "Bill Smith",
    "Bill Smith",
    "2016",
    "<Image>/Py-Astro/Layer Tools/_Stack image layers",
    "RGB, RGB*",
    [
        (PF_SLIDER, "min_bright", "Minimum brightness", 85, (50, 100, 1)),
        (PF_SLIDER, "search_box", "Search box width", 35, (0, 300, 1)),
        (PF_OPTION, "mode"  ,("Layer mode: "), 0 ,["Normal", "Addition", "Screen",]),
        (PF_TOGGLE, "pack", "Flatten image", TRUE),
        (PF_TOGGLE, "auto", "Autocrop image", TRUE),
    ],
    [],
    python_stack_image_layers,
)

main()


