#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python_Stack_Star_Layers 2 - aligns and stacks star images
#   Copyright (C) 2015 Bill Smith
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import math
from gimpfu import *
from array import array

class star_chart:
#
#       star_chart class for finding stars in a photo
#       author - w.smith 2014
#
    def __init__(self, layer, minstar):
#
#   class for identifying stars in a photo
#   author = bill smith (2014)
#
        self.nstars = 0
        self.minstar = minstar
        self.layer = layer
        bpp = layer.bpp
        if bpp <= 4:
            star_factor = 2.56 * float(3 * minstar)
        elif bpp <= 8:
            star_factor = 655.36 * float(3 * minstar)
        else:
            gimp.message("Error - too many bytes per pixel")
            return
        maxpxl = 1000
        width = layer.width
        height = layer.height
        px = [0 for i in range(maxpxl)]
        py = [0 for i in range(maxpxl)]
        idp = [0 for i in range(maxpxl)]
        cc = array("B","\x00"*bpp*maxpxl)
        rgb = array("B", "\x00"*bpp)
        home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
        home_pixels = array("B", home_rgn[0:width, 0:height])
        progress = 0.0
        max_progress = float(width)
        gimp.progress_init("Analysing image " + layer.name + "...")

# find star pixels

        nnn = 0
        for i in range(width):
            for j in range(height):
                k = (i + width * j) * bpp
                rgb = home_pixels[k:k+bpp]
                if bpp <= 4:
                    starbright = float(rgb[0] + rgb[1] + rgb[2])
                elif bpp <= 8:
                    starbright = float((rgb[0] + (rgb[1] << 8)) + (rgb[2] + (rgb[3] << 8)) + (rgb[4] + (rgb[5] << 8)))
                if starbright >= star_factor:
                    px[nnn] = i
                    py[nnn] = j
                    idp[nnn] = nnn
                    mmm=nnn*bpp
                    cc[mmm:mmm+bpp] = rgb
                    nnn += 1
                    if nnn == maxpxl:
                        ee = [0 for n in range(100)]
                        px.extend(ee)
                        py.extend(ee)
                        idp.extend(ee)
                        hh = array("B","\x00"*bpp*100)
                        cc.extend(hh)
                        maxpxl += 100
            progress += 1.0
            gimp.progress_update(progress / max_progress)
        gimp.pdb.gimp_progress_end()
        if nnn == 0:
            gimp.message("Warning - no stars found!")
            return

# group pixels into stars

        progress = 0
        max_progress = 0.5*float(nnn)*float(nnn-1)
        gimp.progress_init("Grouping pixels into stars...")
        self.nstars = nnn
        for i in range(nnn - 1):
            for j in range(i + 1, nnn):
                if abs(px[i] - px[j]) <= 1 and abs(py[i] - py[j]) <= 1:
                    if idp[j] == j:
                        idp[j] = idp[i]
                        self.nstars -= 1
                    elif idp[j] <> idp[i]:
                        p = min(idp[i], idp[j])
                        q = max(idp[i], idp[j])
                        for k in range(nnn):
                            if idp[k] == q:
                                idp[k] = p
                        self.nstars -= 1
                progress += 1.0
                if int(progress) % nnn == 0:
                    gimp.progress_update(progress / max_progress)

        gimp.progress_update(1.0)
        gimp.pdb.gimp_progress_end()
        gimp.message("Number of stars found: "+str(self.nstars))

# sort pixels into star order

        gimp.progress_init("Sorting star pixels ...")
        lst = [i for i in range(nnn)]
        shell_sort(nnn, lst, idp)
        gimp.progress_update(1.0)
        gimp.pdb.gimp_progress_end()

# determine star properties: sizes and positions

        x1 = [0 for i in range(self.nstars)]
        x2 = [0 for i in range(self.nstars)]
        y1 = [0 for i in range(self.nstars)]
        y2 = [0 for i in range(self.nstars)]
        self.sa = [0 for i in range(self.nstars)]
        self.sx = [0 for i in range(self.nstars)]
        self.sy = [0 for i in range(self.nstars)]
        self.sr = [0.0 for i in range(self.nstars)]

        progress = 1
        max_progress = float(nnn)
        gimp.progress_init("Calculating star properties...")

        sn = 0
        pop = 1
        me = idp[0]
        npx = lst[0]
        self.sx[sn] = px[npx]
        self.sy[sn] = py[npx]
        x1[sn] = px[npx]
        x2[sn] = px[npx]
        y1[sn] = py[npx]
        y2[sn] = py[npx]
        for i in range(1,nnn):
            npx = lst[i]
            if idp[i] == me:
                pop += 1
                self.sx[sn] += px[npx]
                self.sy[sn] += py[npx]
                x1[sn] = min(px[npx], x1[sn])
                x2[sn] = max(px[npx], x2[sn])
                y1[sn] = min(py[npx], y1[sn])
                y2[sn] = max(py[npx], y2[sn])
            else:
                self.sx[sn] /= pop
                self.sy[sn] /= pop
                wx = max(x2[sn]-self.sx[sn],self.sx[sn]-x1[sn])
                wy = max(y2[sn]-self.sy[sn],self.sy[sn]-y1[sn])
                self.sr[sn] = math.sqrt(wx*wx+wy*wy)
                self.sa[sn] = math.pi*(wx*wx+wy*wy)
                me = idp[i]
                sn += 1
                pop = 1
                self.sx[sn] = px[npx]
                self.sy[sn] = py[npx]
                x1[sn] = px[npx]
                x2[sn] = px[npx]
                y1[sn] = py[npx]
                y2[sn] = py[npx]
            progress += float(i)
            gimp.progress_update(progress / max_progress)

        self.sx[sn] /= pop
        self.sy[sn] /= pop

        wx = max(x2[sn]-self.sx[sn],self.sx[sn]-x1[sn])
        wy = max(y2[sn]-self.sy[sn],self.sy[sn]-y1[sn])
        self.sr[sn] = math.sqrt(wx*wx+wy*wy)
        self.sa[sn] = math.pi*(wx*wx+wy*wy)

        gimp.pdb.gimp_progress_end()

# sort stars by size (largest first)

        si = [i for i in range(self.nstars)]
        sb = [0 for i in range(self.nstars)]
        sc = [0 for i in range(self.nstars)]
        sd = [0 for i in range(self.nstars)]
        se = [0 for i in range(self.nstars)]
        for i in range(self.nstars):
            sb[i] = self.sa[i]
        shell_sort(self.nstars, si, sb)
        for i in range(self.nstars):
            j = self.nstars - i - 1
            sb[j] = self.sa[si[i]]
            sc[j] = self.sx[si[i]]
            sd[j] = self.sy[si[i]]
            se[j] = self.sr[si[i]]
        self.sa = sb
        self.sx = sc
        self.sy = sd
        self.sr = se

def shell_sort(n, lst, iii):
#
#       method to sort list of integers into ascending order
#       using the shell method
#       author - w.smith 2013
#
    m = n
    while m > 0:
        m = m / 2
        for j in range(n - m):
            i = j
            while i >= 0:
                l = i + m
                if iii[l] < iii[i]:
                    k = lst[i]
                    lst[i] = lst[l]
                    lst[l] = k
                    k = iii[i]
                    iii[i] = iii[l]
                    iii[l] = k
                    i -= m
                else:
                    i = -1

def python_stack_star_layers2(img, layer, arg0, arg1, mode, pack, auto):
    ''' This GIMP plug-in aligns and stacks star images.

    Parameters:
    img        : The currently selected image.
    ref_layer  : The selected (refererence) layer of the current image.
    minstar    : Minimum colour brightness of star objects to select
    search_box : Width of search box for fitting (in pixels)
    mode       : Blending mode for layers
    pack       : Flatten final image option
    auto       : Autocrop finalimage option

    '''
    key = [0, 7, 4]
    minstar = int(arg0)
    if not pack:
        auto = False
    search_box = 2 * (int(arg1) / 2) + 1
    ref_layer = layer
    width = ref_layer.width
    height = ref_layer.height
    num = len(img.layers)
    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)
    pdb.gimp_item_set_visible(ref_layer,True)
    maxprog = float(search_box * search_box)
    ref_chart = star_chart(ref_layer, minstar)
    slim = (search_box / 2 + 1)
    xpos = ypos = xneg = yneg = 0

    for k in range(num):
        mov_layer = img.layers[num - k - 1]
        chkvis = pdb.gimp_item_get_visible(img.layers[num - k - 1])
        if(chkvis == True and mov_layer != ref_layer):

            progress = 0
            mov_chart = star_chart(mov_layer, minstar)
            gimp.progress_init("Aligning image " + mov_layer.name + " ...")

            xdis = 0
            ydis = 0
            overmax = 0.0
            for s in range(slim):
                nlim = max(8 * s, 1)
                for n in range(nlim):
                    progress += 1
                    if progress % 10 == 0:
                        gimp.progress_update(float(progress) / maxprog)
                    if s == 0:
                        x = 0
                        y = 0
                    elif n < 2 * s:
                        x = n - s
                        y = -s
                    elif n < 4 * s:
                        x = s
                        y = n - 3 * s
                    elif n < 6 * s:
                        x = 5 * s - n
                        y = s
                    else:
                        x = -s
                        y = 7 * s - n
                    overlap = 0.0
                    for i in range(min(30, ref_chart.nstars)):
                        for j in range(min(30, mov_chart.nstars)):
                            reach = ref_chart.sr[i] + mov_chart.sr[j]
                            reach2 = math.pow(reach, 2)
                            if (mov_chart.sx[j] >= slim and mov_chart.sx[j] <= width - slim
                                and mov_chart.sy[j] >= slim and mov_chart.sy[j] <= height - slim ):
                                dx = x + mov_chart.sx[j] - ref_chart.sx[i]
                                if abs(dx) <= reach:
                                    dy = y + mov_chart.sy[j] - ref_chart.sy[i]
                                    if abs(dy) <= reach:
                                        d2 = float(dx * dx + dy * dy)
                                        if d2 <= reach2:
                                            if d2 <= math.pow(mov_chart.sr[j] - ref_chart.sr[i], 2):
                                                overlap += min(mov_chart.sa[j], ref_chart.sa[i])
                                            else:
                                                ra = max(mov_chart.sr[j], ref_chart.sr[i])
                                                rb = min(mov_chart.sr[j], ref_chart.sr[i])
                                                dd = math.sqrt(d2)
                                                tt = (d2 + ra * ra - rb * rb) / (2 * dd * ra)
                                                if tt > 1.0:
                                                    tt=1.0
                                                elif tt < -1.0:
                                                    tt=-1.0
                                                ta = math.acos(tt)
                                                tt = (d2 + rb * rb - ra * ra) / (2 * dd * rb)
                                                if tt > 1.0:
                                                    tt=1.0
                                                elif tt < -1.0:
                                                    tt=-1.0
                                                tb = math.acos(tt)
                                                overlap += ra * (ta - math.sin(2 * ta)) + rb * (tb - math.sin(2 * tb))
                    if overlap > overmax:
                        xdis = x
                        ydis = y
                        overmax = overlap

            if xdis <> 0 or ydis <> 0:
                pdb.gimp_layer_translate (mov_layer, xdis, ydis)
                if auto:
                    xpos = max(xpos, xdis)
                    ypos = max(ypos, ydis)
                    xneg = min(xneg, xdis)
                    yneg = min(yneg, ydis)
            gimp.pdb.gimp_progress_end()
            gimp.message("Displacement of "+mov_layer.name+":   "+str(xdis)+"   "+str(ydis))

    if pack:
        n = 0
        for i in range(num):
            if (pdb.gimp_item_get_visible(img.layers[i]) == True):
                n += 1
        k = 0
        for i in range(num):
            if (pdb.gimp_item_get_visible(img.layers[i]) == True):
                if (key[mode] == 0):
                    blend = int(1000 / (n - k)) / 10.0
                else :
                    blend = 100
                pdb.gimp_layer_set_opacity(img.layers[i], blend)
                pdb.gimp_layer_set_mode(img.layers[i], key[mode])
                k += 1
        gimp.progress_init("Flattening image ...")
        img.flatten()
        gimp.pdb.gimp_progress_end()
        if auto:
            new_width = width - xpos + xneg
            new_height = height - ypos + yneg
            pdb.gimp_image_crop(img, new_width, new_height, xpos, ypos)
        pdb.gimp_item_set_name(img.layers[0],"Stacked image")
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()


def abs(x):
    if x < 0:
         x = -x
    return x

def min(x,y):
    if x < y:
        z = x
    else:
        z = y
    return z

def max(x,y):
    if x > y:
        z = x
    else:
        z = y
    return z

register(
    "python_stack_star_layers",
    "Align and stack star images",
    "Align and stack star images - shifting only (no rotation)",
    "Bill Smith",
    "Bill Smith",
    "2014",
    "<Image>/Py-Astro/Layer Tools/_Stack star layers ...",
    "RGB, RGB*",
    [
        (PF_SLIDER, "minstar", "Star minimum", 85, (50, 100, 1)),
        (PF_SLIDER, "search_box", "Search box width", 35, (0, 300, 1)),
        (PF_OPTION, "mode"  ,("Layer mode: "), 0 ,["Normal", "Addition", "Screen",]),
        (PF_TOGGLE, "pack", "Flatten image", TRUE),
        (PF_TOGGLE, "auto", "Autocrop image", TRUE),
    ],
    [],
    python_stack_star_layers2,
)

main()


