#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python_Stack_Sun_Layers - aligns and stacks Sun images
#   Copyright (C) 2014 Bill Smith
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import math
from gimpfu import *
from array import array

class sun_chart:
#
#       sun_chart class for finding the sun in a photo
#       author - w.smith 2014
#
    def __init__(self, layer, minsun):
#
#   class for identifying the sun in a photo
#   author = bill smith (2014)
#
        self.minsun = minsun
        self.layer = layer
        maxpxl = 1000
        bpp = layer.bpp
        if bpp <= 4:
            sun_factor = 2.56 * float(3 * minsun)
        elif bpp <= 8:
            sun_factor = 655.36 * float(3 * minsun)
        else:
            gimp.message("Error - too many bytes per pixel")
            return
        width = layer.width
        height = layer.height
        px = [0 for i in range(maxpxl)]
        py = [0 for i in range(maxpxl)]
        cc = array("B","\x00"*bpp*maxpxl)
        rgb = array("B", "\x00"*bpp)
        home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
        home_pixels = array("B", home_rgn[0:width, 0:height])
        progress = 0.0
        max_progress = float(width * height)
        gimp.progress_init("Analysing file " + layer.name + "...")

# find sun pixels

        nnn = 0
        for i in range(width):
            for j in range(height):
                k = (i + width * j) * bpp
                rgb = home_pixels[k:k+bpp]
                if bpp <= 4:
                    sunbright = float(rgb[0] + rgb[1] + rgb[2])
                elif bpp <= 8:
                    sunbright = float((rgb[0] + (rgb[1] << 8)) + (rgb[2] + (rgb[3] << 8)) + (rgb[4] + (rgb[5] << 8)))
                if sunbright > sun_factor:
                    px[nnn] = i
                    py[nnn] = j
                    mmm=nnn*bpp
                    cc[mmm:mmm+bpp] = rgb
                    nnn += 1
                    if nnn == maxpxl:
                        ee = [0 for n in range(100)]
                        px.extend(ee)
                        py.extend(ee)
                        hh = array("B","\x00"*bpp*100)
                        cc.extend(hh)
                        maxpxl += 100
            progress += float(width)
            gimp.progress_update(progress / max_progress)

        gimp.pdb.gimp_progress_end()
        if nnn == 0:
            gimp.message("Warning - no sun found!")

        gimp.message("Number of Sun pixels found: "+str(nnn))

# determine sun properties: size, colour and position

        self.sa = 0
        self.sx = 0
        self.sy = 0
        self.sr = 0
        self.co = array("B","\x00"*bpp)
        bgr = array("B", "\x00"*bpp)

        progress = 1
        max_progress = float(nnn)
        gimp.progress_init("Calculating sun properties...")

        rgb = cc[0:0+bpp]
        aa = 0.0
        if bpp <= 4:
            rr = float(rgb[0])
            gg = float(rgb[1])
            bb = float(rgb[2])
            if bpp > 3:
                aa = float(rgb[3])
        elif bpp <= 8:
            rr = float(rgb[0]+(rgb[1] << 8))
            gg = float(rgb[2]+(rgb[3] << 8))
            bb = float(rgb[4]+(rgb[5] << 8))
            if bpp > 6:
                aa = float(rgb[6]+(rgb[7] << 8))
        self.sx = px[0]
        self.sy = py[0]
        for i in range(1,nnn):
            uuu = i*bpp
            rgb = cc[uuu:uuu+bpp]
            if bpp <= 4:
                rr += float(rgb[0])
                gg += float(rgb[1])
                bb += float(rgb[2])
                if bpp > 3:
                    aa += float(rgb[3])
            elif bpp <= 8:
                rr += float(rgb[0]+(rgb[1] << 8))
                gg += float(rgb[2]+(rgb[3] << 8))
                bb += float(rgb[4]+(rgb[5] << 8))
                if bpp > 6:
                    aa += float(rgb[6]+(rgb[7] << 8))
            self.sx += px[i]
            self.sy += py[i]
            progress += float(i)
            gimp.progress_update(progress / max_progress)

        self.sa = nnn
        self.sx /= nnn
        self.sy /= nnn
        self.sr = math.sqrt(nnn / math.pi)
        top = max(bb,gg)
        top = max(top,rr)
        if bpp <= 4:
            bgr[0] = int((255.0 * (rr / nnn)) / top)
            bgr[1] = int((255.0 * (gg / nnn)) / top)
            bgr[2] = int((255.0* (bb / nnn)) / top)
            if bpp > 3:
                bgr[3] = int(aa / nnn)
        elif bpp <= 8:
            rr = (65535 * (rr / nnn)) / top
            gg = (65535 * (gg / nnn)) / top
            bb = (65535 * (bb / nnn)) / top
            uu = int(rr)
            bgr[1] = (uu >> 8)
            bgr[0] = uu - (bgr[1] << 8)
            uu = int(gg)
            bgr[3] = (uu >> 8)
            bgr[2] = uu - (bgr[3] << 8)
            uu = int(bb)
            bgr[5] = (uu >> 8)
            bgr[4] = uu - (bgr[5] << 8)
            if bpp > 6:
                uu = int(bb)
                bgr[7] = (uu >> 8)
                bgr[6] = uu - (bgr[7] << 8)
        self.co[0:0+bpp] = bgr

        gimp.pdb.gimp_progress_end()

def python_stack_sun_layers(img, layer, arg0, arg1, pack, auto):
    ''' This GIMP plug-in aligns and stacks sun images.

    Parameters:
    img        : The currently selected image.
    ref_layer  : The selected (refererence) layer of the current image.
    minsun     : Minimum colour brightness of sun objects to select
    search_box : Width of search box for fitting (in pixels)
    pack       : Flatten final image option
    auto       : Autocrop finalimage option

    '''

    minsun = int(arg0)
    if not pack:
        auto = False
    search_box = 2 * (int(arg1) / 2) + 1
    ref_layer = layer
    width = ref_layer.width
    height = ref_layer.height
    num = len(img.layers)
    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)
    maxprog = float(search_box * search_box)
    ref_chart = sun_chart(ref_layer, minsun)
    slim = (search_box / 2 + 1)
    xpos = ypos = xneg = yneg = 0

    for k in range(num):
        mov_layer = img.layers[num - k - 1]
        chkvis = pdb.gimp_item_get_visible(img.layers[num - k - 1])
        if(chkvis == True and mov_layer != ref_layer):

            progress = 0
            mov_chart = sun_chart(mov_layer, minsun)
            gimp.progress_init("Aligning layer " + mov_layer.name + " ...")

            xdis = 0
            ydis = 0
            overmax = 0.0
            for s in range(slim):
                nlim = max(8 * s, 1)
                for n in range(nlim):
                    progress += 1
                    if progress % 10 == 0:
                        gimp.progress_update(float(progress) / maxprog)
                    if s == 0:
                        x = 0
                        y = 0
                    elif n < 2 * s:
                        x = n - s
                        y = -s
                    elif n < 4 * s:
                        x = s
                        y = n - 3 * s
                    elif n < 6 * s:
                        x = 5 * s - n
                        y = s
                    else:
                        x = -s
                        y = 7 * s - n
                    overlap = 0.0
                    reach = ref_chart.sr + mov_chart.sr
                    reach2 = math.pow(reach, 2)
                    if (mov_chart.sx >= slim and mov_chart.sx <= width - slim
                        and mov_chart.sy >= slim and mov_chart.sy <= height - slim ):
                        dx = x + mov_chart.sx - ref_chart.sx
                        if abs(dx) <= reach:
                            dy = y + mov_chart.sy - ref_chart.sy
                            if abs(dy) <= reach:
                                d2 = float(dx * dx + dy * dy)
                                if d2 <= reach2:
                                    if d2 <= math.pow(mov_chart.sr - ref_chart.sr, 2):
                                        overlap += min(mov_chart.sa, ref_chart.sa)
                                    else:
                                        ra = max(mov_chart.sr, ref_chart.sr)
                                        rb = min(mov_chart.sr, ref_chart.sr)
                                        dd = math.sqrt(d2)
                                        ta = math.acos((d2 + ra * ra - rb * rb) / (2 * dd * ra))
                                        tb = math.acos((d2 + rb * rb - ra * ra) / (2 * dd * rb))
                                        overlap += ra * (ta - math.sin(2 * ta)) + rb * (tb - math.sin(2 * tb))
                    if overlap > overmax:
                        xdis = x
                        ydis = y
                        overmax = overlap

            if xdis <> 0 or ydis <> 0:
                pdb.gimp_layer_translate (mov_layer, xdis, ydis)
                if auto:
                    xpos = max(xpos, xdis)
                    ypos = max(ypos, ydis)
                    xneg = min(xneg, xdis)
                    yneg = min(yneg, ydis)
            gimp.pdb.gimp_progress_end()

    if pack:
        n = 0
        for i in range(num):
            if (pdb.gimp_item_get_visible(img.layers[i]) == True):
                n += 1
        k = 0
        for i in range(num):
            if (pdb.gimp_item_get_visible(img.layers[i]) == True):
                blend = int(1000 / (n - k)) / 10.0
                pdb.gimp_layer_set_opacity(img.layers[i], blend)
                pdb.gimp_layer_set_mode(img.layers[i], 0)
                k += 1
        img.flatten()
        if auto:
            new_width = width - xpos + xneg
            new_height = height - ypos + yneg
            pdb.gimp_image_crop(img, new_width, new_height, xpos, ypos)
        pdb.gimp_item_set_name(img.layers[0],"Stacked image")
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()


def abs(x):
    if x < 0:
         x = -x
    return x

def min(x,y):
    if x < y:
        z = x
    else:
        z = y
    return z

def max(x,y):
    if x > y:
        z = x
    else:
        z = y
    return z

register(
    "python_stack_sun_layers",
    "Aligns and stacks sun images",
    "Aligns and stacks sun images - shifting only (no rotation)",
    "Bill Smith",
    "Bill Smith",
    "2014",
    "<Image>/Py-Astro/Layer Tools/_Stack sun layers ...",
    "RGB, RGB*",
    [
        (PF_SLIDER, "minsun", "Sun minimum", 60, (50, 100, 1)),
        (PF_SLIDER, "search_box", "Search box width", 35, (0, 300, 1)),
        (PF_TOGGLE, "pack", "Flatten image", TRUE),
        (PF_TOGGLE, "auto", "Autocrop image", TRUE),
    ],
    [],
    python_stack_sun_layers,
)

main()


