#!/usr/bin/env python
#   Gimp-Python - allows the writing of Gimp plugins in Python.
#
#   Python_Star_Halo - apply halo to star images!
#   Copyright (C) 2014 Bill Smith
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import math
from gimpfu import *
from array import array

class star_chart:
#
#       star_chart class for finding stars in a photo
#       author - w.smith 2014
#
    def __init__(self, layer, minstar):
#
#   class for identifying stars in a photo
#   author = bill smith (2014)
#
        self.nstars = 0
        self.minstar = minstar
        self.layer = layer
        maxpxl = 1000
        bpp = layer.bpp
        if bpp <= 4:
            star_factor = 2.56 * float(3 * minstar)
        elif bpp <= 8:
            star_factor = 655.36 * float(3 * minstar)
        width = layer.width
        height = layer.height
        px = [0 for i in range(maxpxl)]
        py = [0 for i in range(maxpxl)]
        idp = [0 for i in range(maxpxl)]
        cc = array("B","\x00"*bpp*maxpxl)
        rgb = array("B", "\x00"*bpp)
        home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
        home_pixels = array("B", home_rgn[0:width, 0:height])
        progress = 0.0
        max_progress = float(width * height)
        gimp.progress_init("Finding star pixels...")

# find star pixels

        nnn = 0
        for i in range(width):
            for j in range(height):
                k = (i + width * j) * bpp
                rgb = home_pixels[k:k+bpp]
                if bpp <= 4:
                    starbright = float(rgb[0] + rgb[1] + rgb[2])
                elif bpp <= 8:
                    starbright = float((rgb[0] + (rgb[1] << 8)) + (rgb[2] + (rgb[3] << 8)) + (rgb[4] + (rgb[5] << 8)))
                if starbright >= star_factor:
                    px[nnn] = i
                    py[nnn] = j
                    idp[nnn] = nnn
                    mmm=nnn*bpp
                    cc[mmm:mmm+bpp] = rgb
                    nnn += 1
                    if nnn == maxpxl:
                        ee = [0 for n in range(100)]
                        px.extend(ee)
                        py.extend(ee)
                        idp.extend(ee)
                        hh = array("B","\x00"*bpp*100)
                        cc.extend(hh)
                        maxpxl += 100
            progress += float(width)
            gimp.progress_update(progress / max_progress)

        gimp.pdb.gimp_progress_end()
        if nnn == 0:
            gimp.message("Warning - no stars found!")
            return

# group pixels into stars

        progress = 0
        max_progress = 0.5*float(nnn)*float(nnn-1)
        gimp.progress_init("Grouping pixels into stars...")
        self.nstars = nnn
        for i in range(nnn - 1):
            for j in range(i + 1, nnn):
                if abs(px[i] - px[j]) <= 1 and abs(py[i] - py[j]) <= 1:
                    if idp[j] == j:
                        idp[j] = idp[i]
                        self.nstars -= 1
                    elif idp[j] <> idp[i]:
                        p = min(idp[i], idp[j])
                        q = max(idp[i], idp[j])
                        for k in range(nnn):
                            if idp[k] == q:
                                idp[k] = p
                        self.nstars -= 1
                progress += 1.0
                if int(progress) % nnn == 0 :
                    gimp.progress_update(progress / max_progress)

        gimp.progress_update(1.0)
        gimp.pdb.gimp_progress_end()
        gimp.message("Number of stars found: "+str(self.nstars))

# sort pixels into star order

        gimp.progress_init("Sorting star pixels ...")
        lst = [i for i in range(nnn)]
        shell_sort(nnn, lst, idp)
        gimp.progress_update(1.0)
        gimp.pdb.gimp_progress_end()

# determine star properties: sizes, colours and positions

        self.sa = [0 for i in range(self.nstars)]
        self.sx = [0 for i in range(self.nstars)]
        self.sy = [0 for i in range(self.nstars)]
        self.co = array("B","\x00"*bpp*self.nstars)
        bgr = array("B", "\x00"*bpp)

        progress = 1
        max_progress = float(nnn)
        gimp.progress_init("Calculating star properties...")

        sn = 0
        pop = 1
        me = idp[0]
        npx = lst[0]
        uuu = npx*bpp
        rgb = cc[uuu:uuu+bpp]
        if bpp <= 4:
            rr = (rgb[0])
            gg = (rgb[1])
            bb = (rgb[2])
            if bpp > 3:
                aa = int(rgb[3])
            else:
                aa = 0
        else:
            rr = (rgb[0]+(rgb[1] << 8))
            gg = (rgb[2]+(rgb[3] << 8))
            bb = (rgb[4]+(rgb[5] << 8))
            if bpp > 6:
                aa = int(rgb[6]+(rgb[7] << 8))
            else:
                aa = 0
        self.sx[sn] = px[npx]
        self.sy[sn] = py[npx]
        for i in range(1,nnn):
            npx = lst[i]
            uuu = npx*bpp
            rgb = cc[uuu:uuu+bpp]
            if idp[i] == me:
                pop += 1
                if bpp <= 4:
                    rr += (rgb[0])
                    gg += (rgb[1])
                    bb += (rgb[2])
                    if bpp > 3:
                        aa += int(rgb[3])
                else:
                    rr += (rgb[0]+(rgb[1] << 8))
                    gg += (rgb[2]+(rgb[3] << 8))
                    bb += (rgb[4]+(rgb[5] << 8))
                    if bpp > 6:
                        aa += int(rgb[6]+(rgb[7] << 8))
                self.sx[sn] += px[npx]
                self.sy[sn] += py[npx]
            else:
                self.sa[sn] = math.sqrt(float(pop) / math.pi) + 1
                self.sx[sn] /= pop
                self.sy[sn] /= pop
                if bpp <= 4:
                    bgr[0] = rr / pop
                    bgr[1] = gg / pop
                    bgr[2] = bb / pop
                    if bpp > 3:
                       bgr[3] = aa / pop
                else:
                    mm = rr / pop
                    bgr[1] = (mm >> 8)
                    bgr[0] = mm - (bgr[1] << 8)
                    mm = gg /pop
                    bgr[3] = (mm >> 8)
                    bgr[2] = mm - (bgr[3] << 8)
                    mm = bb / pop
                    bgr[5] = (mm >> 8)
                    bgr[4] = mm - (bgr[5] << 8)
                    if bpp > 6:
                        mm = aa / pop
                        bgr[7] = (mm >> 8)
                        bgr[6] = mm - (bgr[7] << 8)
                vvv = sn*bpp
                self.co[vvv:vvv+bpp] = bgr
                me = idp[i]
                sn += 1
                pop = 1
                if bpp <= 4:
                    rr = (rgb[0])
                    gg = (rgb[1])
                    bb = (rgb[2])
                    if bpp > 3:
                        aa = int(rgb[3])
                    else:
                        aa = 0
                else:
                    rr = (rgb[0]+(rgb[1] << 8))
                    gg = (rgb[2]+(rgb[3] << 8))
                    bb = (rgb[4]+(rgb[5] << 8))
                    if bpp > 6:
                        aa = int(rgb[6]+(rgb[7] << 8))
                    else:
                        aa = 0
                self.sx[sn] = px[npx]
                self.sy[sn] = py[npx]
            progress += float(i)
            gimp.progress_update(progress / max_progress)

        self.sa[sn] = math.sqrt(float(pop) / math.pi) + 1
        self.sx[sn] /= pop
        self.sy[sn] /= pop
        if bpp <= 4:
            bgr[0] = rr / pop
            bgr[1] = gg / pop
            bgr[2] = bb / pop
            if bpp > 3:
                bgr[3] = aa / pop
        else:
            mm = rr / pop
            bgr[1] = (mm >> 8)
            bgr[0] = mm - (bgr[1] << 8)
            mm = gg /pop
            bgr[3] = (mm >> 8)
            bgr[2] = mm - (bgr[3] << 8)
            mm = bb / pop
            bgr[5] = (mm >> 8)
            bgr[4] = mm - (bgr[5] << 8)
            if bpp > 6:
                mm = aa / pop
                bgr[7] = (mm >> 8)
                bgr[6] = mm - (bgr[7] << 8)
        vvv = sn*bpp
        self.co[vvv:vvv+bpp] = bgr
        gimp.pdb.gimp_progress_end()

def shell_sort(n, lst, iii):
#
#       method to sort list of integers into ascending order
#       using the shell method
#       author - w.smith 2013
#
    m = n
    while m > 0:
        m = m / 2
        for j in range(n - m):
            i = j
            while i >= 0:
                l = i + m
                if iii[l] < iii[i]:
                    k = lst[i]
                    lst[i] = lst[l]
                    lst[l] = k
                    k = iii[i]
                    iii[i] = iii[l]
                    iii[l] = k
                    i -= m
                else:
                    i = -1

def python_star_halo(img, layer, arg0, arg1, arg2, arg3):
    ''' This GIMP plug-in draws halos around star images.

    Parameters:
    img       : The currently selected image.
    layer     : The layer of the currently selected image.
    minstar   : Minimum colour brightness of star objects
    size      : The size (in pixels) of the effect
    bright    : Brightness factor of halo
    key       : The key indicating which star halo is required
              = 0 for disc
              = 1 for plus
              = 2 for cross
              = 3 for asterisk
    '''
    precision = pdb.gimp_image_get_precision(img)
    bpp = layer.bpp
    if bpp > 8:
        gimp.message("Error - too many bytes per pixel")
        return
    minstar = int(arg0)
    size = float(arg1)
    bright = float(arg2)
    key = int(arg3)

    halo = ['halo', 'plus', 'cross', 'asterisk']

    width = layer.width
    height = layer.height
    rgb = array("B", "\x00"*bpp)
    gbr = array("B", "\x00"*bpp)

    gimp.context_push()
    pdb.gimp_image_undo_group_start(img)
    progress = 0
    max_progress = width * height

    chart = star_chart(layer, minstar)
    gimp.progress_init("Drawing star halos ...")
    home_rgn = layer.get_pixel_rgn(0, 0, width, height, False, False)
    home_pixels = array("B", home_rgn[0:width, 0:height])
    newlay = gimp.Layer(img, layer.name + "_"+halo[key], width, height,
                        layer.type, layer.opacity, layer.mode)
    img.add_layer(newlay, -1)
    dest_rgn = newlay.get_pixel_rgn(0, 0, width, height, True, True)
    dest_pixels = array("B", home_rgn[0:width, 0:height])


    for n in range(chart.nstars):
        nnn = int(chart.sa[n] * size)
        if key == 0 :
            (npoints, xxx, yyy, mask) = get_halo(nnn)
        elif key == 1:
            (npoints, xxx, yyy, mask) = get_plus(nnn)
        elif key == 2:
            (npoints, xxx, yyy, mask) = get_cross(nnn)
        elif key == 3:
            (npoints, xxx, yyy, mask) = get_asterisk(nnn)
        nrm = bright * npoints
        m=max(min(int(chart.sa[n]),5),1)
        for i in range(-m,m+1):
            for j in range(-m, m+1):
                if math.pi * float(i * i + j * j) <= float(m * m):
                    u = chart.sx[n] + i
                    v = chart.sy[n] + j
                    if u >= 0 and u < width and v >= 0 and v < height:
                        uv = (u + width * v) * bpp
                        rgb = home_pixels[uv:uv+bpp]
                        for k in range(npoints):
                            p = u + xxx[k]
                            q = v + yyy[k]
                            if p >= 0 and p < width and q >= 0 and q < height:
                                pq = (p + width * q) * bpp
                                bgr = dest_pixels[pq:pq+bpp]
                                if bpp <= 4 :
                                    rr = int((bgr[0]) + (nrm * mask[k] * rgb[0]) / 5)
                                    gg = int((bgr[1]) + (nrm * mask[k] * rgb[1]) / 5)
                                    bb = int((bgr[2]) + (nrm * mask[k] * rgb[2]) / 5)
                                    bgr[0] = min(255, max(0, rr))
                                    bgr[1] = min(255, max(0, gg))
                                    bgr[2] = min(255, max(0, bb))
                                else :
                                    rr = int((bgr[0]+(bgr[1] << 8)) + (nrm * mask[k] * (rgb[0]+(rgb[1] << 8))) / 5)
                                    gg = int((bgr[2]+(bgr[3] << 8)) + (nrm * mask[k] * (rgb[2]+(rgb[3] << 8))) / 5)
                                    bb = int((bgr[4]+(bgr[5] << 8)) + (nrm * mask[k] * (rgb[4]+(rgb[5] << 8))) / 5)
                                    mm = min(65535, max(0, rr))
                                    bgr[1] = (mm >> 8)
                                    bgr[0] = mm - (bgr[1] << 8)
                                    mm = min(65535, max(0, gg))
                                    bgr[3] = (mm >> 8)
                                    bgr[2] = mm - (bgr[3] << 8)
                                    mm = min(65535, max(0, bb))
                                    bgr[5] = (mm >> 8)
                                    bgr[4] = mm - (bgr[5] << 8)
                                dest_pixels[pq:pq+bpp] = bgr

        progress += 1
        gimp.progress_update(float(progress) / chart.nstars)
    dest_rgn[0:width, 0:height] = dest_pixels.tostring()
    newlay.flush()
    newlay.merge_shadow(True)
    newlay.update(0, 0, width, height)
    pdb.gimp_image_undo_group_end(img)
    gimp.context_pop()
    gimp.pdb.gimp_progress_end()

def get_halo(nnn):
#
#   method to construct a halo effect
#   author = bill smith (2014)
#
    size = 2 * nnn + 1
    npoints = size * size
    xxx = [0 for i in range(npoints)]
    yyy = [0 for i in range(npoints)]
    mask = [0 for i in range(npoints)]
    kkk = 0
    nrm = 0.0
    for i in range(-nnn, nnn+1):
     for j in range(-nnn, nnn+1):
        if i * i + j * j <= nnn * nnn:
            rrr = math.sqrt(float(i * i + j * j)) / nnn
            arg = (0.75 * math.pi * rrr)
            xxx[kkk] = i
            yyy[kkk] = j
            if i == j and j == 0:
                mask[kkk] = 1.0
            else:
                mask[kkk] = math.sin(arg) / arg
            nrm += mask[kkk]
            kkk += 1

    npoints = kkk
    for i in range(npoints):
        mask[i] /= (20.0 * nrm)

    return npoints, xxx, yyy, mask

def get_plus(nnn):
#
#   method to construct a plus effect
#   author = bill smith (2014)
#
    size = 2 * nnn + 1
    npoints = 2 * size - 1
    xxx = [0 for i in range(npoints)]
    yyy = [0 for i in range(npoints)]
    mask  = [0 for i in range(npoints)]
    kkk = 0
    xxx[kkk] =0
    yyy[kkk] = 0
    fff = float(nnn)
    mask[kkk] = fff
    nrm = fff
    kkk += 1
    for i in range(1 ,nnn+1):
        arg = (0.75 * math.pi * i) / nnn
        fff = math.sin(arg) / arg
        xxx[kkk] = i
        yyy[kkk] = 0
        mask[kkk] = fff
        nrm += fff
        kkk += 1
        xxx[kkk] = -i
        yyy[kkk] = 0
        mask[kkk] = fff
        nrm += fff
        kkk += 1
        xxx[kkk] = 0
        yyy[kkk] = i
        mask[kkk] = fff
        nrm += fff
        kkk += 1
        xxx[kkk] = 0
        yyy[kkk] = -i
        mask[kkk] = fff
        nrm += fff
        kkk += 1

    for i in range(npoints):
        mask[i] /= (5.0 * nrm)
    return npoints, xxx, yyy, mask

def get_cross(nnn):
#
#   method to construct a cross effect
#   author = bill smith (2014)
#
    size = 2 * nnn + 1
    npoints = 2 * size - 1
    xxx = [0 for i in range(npoints)]
    yyy = [0 for i in range(npoints)]
    mask = [0 for i in range(npoints)]
    kkk = 0
    fff = 1.0
    xxx[kkk] = 0
    yyy[kkk] = 0
    mask[kkk] = fff
    nrm = fff
    kkk += 1
    for i in range(1, nnn+1):
        arg = (0.75 * math.pi * i) / nnn
        fff = math.sin(arg) / arg
        xxx[kkk] = i
        yyy[kkk] = i
        mask[kkk] = fff
        nrm += fff
        kkk += 1
        xxx[kkk] = -i
        yyy[kkk] = -i
        mask[kkk] = fff
        nrm += fff
        kkk += 1
        xxx[kkk] = -i
        yyy[kkk] = i
        mask[kkk] = fff
        nrm += fff
        kkk += 1
        xxx[kkk] = i
        yyy[kkk] = -i
        mask[kkk] = fff
        nrm += fff
        kkk += 1

    for i in range(npoints):
        mask[i] /= (5.0 * nrm)
    return npoints, xxx, yyy, mask

def get_asterisk(nnn):
#
#   method to construct an asterisk effect
#   author = bill smith (2014)
#
    size = 2 * nnn + 1
    npoints = 4 * nnn + 4 * (nnn / 3) + 1
    xxx = [0 for i in range(npoints)]
    yyy = [0 for i in range(npoints)]
    mask = [0 for i in range(npoints)]
    kkk = 0
    fff = 1.0
    xxx[kkk] = 0
    yyy[kkk] = 0
    mask[kkk] = fff
    nrm = fff
    kkk += 1
    for i in range(1, nnn+1):
        arg = (0.75 * math.pi * i) / nnn
        fff = math.sin(arg) / arg
        xxx[kkk] = i
        yyy[kkk] = 0
        mask[kkk] = fff
        nrm += fff
        kkk += 1
        xxx[kkk] = -i
        yyy[kkk] = 0
        mask[kkk] = fff
        nrm += fff
        kkk += 1
        xxx[kkk] = 0
        yyy[kkk] = i
        mask[kkk] = fff
        nrm += fff
        kkk += 1
        xxx[kkk] = 0
        yyy[kkk] = -i
        mask[kkk] = fff
        nrm += fff
        kkk += 1
        if i <= nnn / 3:
            xxx[kkk] = i
            yyy[kkk] = i
            mask[kkk] = fff
            nrm += fff
            kkk += 1
            xxx[kkk] = -i
            yyy[kkk] = -i
            mask[kkk] = fff
            nrm += fff
            kkk += 1
            xxx[kkk] = -i
            yyy[kkk] = i
            mask[kkk] = fff
            nrm += fff
            kkk += 1
            xxx[kkk] = i
            yyy[kkk] = -i
            mask[kkk] = fff
            nrm += fff
            kkk += 1

    for i in range(npoints):
        mask[i] /= (5.0 * nrm)

    return npoints, xxx, yyy, mask

def abs(x):
    if x < 0:
         x = -x
    return x

def min(x,y):
    if x < y:
        z = x
    else:
        z = y
    return z

def max(x,y):
    if x > y:
        z = x
    else:
        z = y
    return z

register(
    "python_star_halo",
    "Draws halos around a star image",
    "Augments a star image with disc halos or diffraction spikes",
    "Bill Smith",
    "Bill Smith",
    "2014",
    "<Image>/Py-Astro/Effects/_Star halos ...",
    "RGB, RGB*",
    [
        (PF_SLIDER, "minstar", "Star minimum", 85, (20, 100, 1)),
        (PF_SLIDER, "size", "Halo radius", 2.5, (1.0, 10.0, 0.1)),
        (PF_SLIDER, "bright", "Halo brightness", 1.5, (1.0, 5.0, 0.1)),
        (PF_OPTION, "key"  ,("Star halo: "), 0 ,["Halo", "Plus", "Cross", "Asterisk"]),
    ],
    [],
    python_star_halo,
)

main()


